"""Test GLUT forward-compatible mode..."""

from __future__ import print_function
import OpenGL

OpenGL.FORWARD_COMPATIBLE_ONLY = True
OpenGL.ERROR_CHECKING = True
# OpenGL.USE_ACCELERATE = False
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
import time

start = time.time()

from OpenGL.GL.AMD.debug_output import glGetDebugMessageLogAMD
from OpenGL.GL.ARB.debug_output import glGetDebugMessageLogARB
from OpenGL.GL.KHR.debug import glGetDebugMessageLogKHR
from OpenGL.GL.VERSION.GL_4_3 import glGetDebugMessageLog

glGetDebugMessageLog = extensions.alternate(
    glGetDebugMessageLog,
    glGetDebugMessageLogARB,
    glGetDebugMessageLogKHR,
    glGetDebugMessageLogAMD,
)

window = None


def display():
    try:
        glutSetWindow(window)
        glClearColor(0.0, 0.0, (time.time() % 1.0) / 1.0, 0.0)
        glClear(GL_COLOR_BUFFER_BIT)
        try:
            glGetString(GL_EXTENSIONS)
        except GLError:
            pass
        else:
            print('Egads, glGetString should not have worked!')
        assert bool(glGenVertexArrays), "Should have vertex array support in 3.2"
        for message in get_debug_messages():
            print(message)
        glFlush()
        glutSwapBuffers()
    except Exception:
        glutDestroyWindow(window)
        raise


def get_debug_messages():
    messages = []
    count = glGetIntegerv(GL_DEBUG_LOGGED_MESSAGES)
    max_size = int(glGetIntegerv(GL_MAX_DEBUG_MESSAGE_LENGTH))
    source = GLenum()
    type = GLenum()
    id = GLenum()
    severity = GLenum()
    length = GLsizei()
    buffer = ctypes.create_string_buffer(max_size)
    for i in range(count):
        result = glGetDebugMessageLog(
            1, max_size, source, type, id, severity, length, buffer
        )
        if result:
            messages.append(
                {
                    'message': buffer[: length.value],
                    'type': type.value,
                    'id': id.value,
                    'severity': severity.value,
                    'source': source.value,
                }
            )
    return messages
    assert len(messages), messages


size = (250, 250)


def reshape(*args):
    global size
    size = args
    glViewport(*((0, 0) + args))
    display()


def printFunction(name):
    def onevent(*args):
        print('%s -> %s' % (name, ", ".join([str(a) for a in args])))

    return onevent


def idle(*args, **named):
    sys.stdout.write("OK\n")
    sys.stdout.flush()
    global window
    glutDestroyWindow(window)
    try:
        if fgDeinitialize:
            fgDeinitialize(False)
    except NameError:
        pass  # Older PyOpenGL, you may see a seg-fault here...

    os._exit(0)


if __name__ == "__main__":
    import sys

    newArgv = glutInit(sys.argv)
    glutInitContextVersion(3, 1)
    glutInitContextFlags(GLUT_FORWARD_COMPATIBLE | GLUT_DEBUG)
    glutInitContextProfile(GLUT_CORE_PROFILE)
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH)

    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS)
    glutInitWindowSize(250, 250)
    glutInitWindowPosition(100, 100)
    window = glutCreateWindow("hello")
    print('window', repr(window))
    glutDisplayFunc(display)
    glutReshapeFunc(reshape)
    glutMouseFunc(printFunction('Mouse'))
    glutEntryFunc(printFunction('Entry'))
    glutKeyboardFunc(printFunction('Keyboard'))
    glutKeyboardUpFunc(printFunction('KeyboardUp'))
    glutMotionFunc(printFunction('Motion'))
    glutIdleFunc(idle)
    glutPassiveMotionFunc(printFunction('PassiveMotion'))
    glutVisibilityFunc(printFunction('Visibility'))
    glutWindowStatusFunc(printFunction('WindowStatus'))
    glutSpecialFunc(printFunction('Special'))
    glutSpecialUpFunc(printFunction('SpecialUp'))

    glutMainLoop()



import basetestcase
import os, sys
import OpenGL
from OpenGL.arrays import arraydatatype
from OpenGL.GL import *

try:
    import numpy as np
except ImportError as err:
    np = None
HERE = os.path.abspath(os.path.dirname(__file__))
import pytest
from OpenGL import acceleratesupport


class TestCoreDatatype(basetestcase.BaseTest):
    def test_arrayPointer(self):
        dt = arraydatatype.GLuintArray
        d = dt.zeros((3,))
        dp = dt.typedPointer(d)
        assert dp[0] == 0
        assert dp[1] == 0
        assert dp[2] == 0
        dp[1] = 1
        assert dp[1] == 1
        assert d[1] == 1

    def test_ctypes_array(self):
        color = (GLfloat * 3)(0, 1, 0)
        glColor3fv(color)

    if not OpenGL.ERROR_ON_COPY:

        def test_pointers(self):
            """Test that basic pointer functions work"""
            vertex = GLdouble * 3
            vArray = vertex * 2
            glVertexPointerd([[2, 3, 4, 5], [2, 3, 4, 5]])
            glVertexPointeri(([2, 3, 4, 5], [2, 3, 4, 5]))
            glVertexPointers([[2, 3, 4, 5], [2, 3, 4, 5]])
            glVertexPointerd(vArray(vertex(2, 3, 4), vertex(2, 3, 4)))
            myVector = vArray(vertex(2, 3, 4), vertex(2, 3, 4))
            glVertexPointer(
                3, GL_DOUBLE, 0, ctypes.cast(myVector, ctypes.POINTER(GLdouble))
            )

            repr(glVertexPointerb([[2, 3], [4, 5]]))
            glVertexPointerf([[2, 3], [4, 5]])
            assert arrays.ArrayDatatype.dataPointer(None) == None
            glVertexPointerf(None)

            glNormalPointerd([[2, 3, 4], [2, 3, 4]])
            glNormalPointerd(None)

            glTexCoordPointerd([[2, 3, 4], [2, 3, 4]])
            glTexCoordPointerd(None)

            glColorPointerd([[2, 3, 4], [2, 3, 4]])
            glColorPointerd(None)

            glEdgeFlagPointerb([0, 1, 0, 0, 1, 0])
            glEdgeFlagPointerb(None)

            glIndexPointerd([0, 1, 0, 0, 1, 0])
            glIndexPointerd(None)

            glColor4fv([0, 0, 0, 1])

            # string data-types...
            import struct

            s = struct.pack('>iiii', 2, 3, 4, 5) * 2
            glVertexPointer(4, GL_INT, 0, s)

        def test_texture(self):
            """Test texture (requires OpenGLContext and PIL)"""
            try:
                from OpenGLContext import texture
                import Image
                from OpenGL.GLUT import glutSolidTeapot
            except ImportError:
                pass
            else:
                assert glutSolidTeapot
                glEnable(GL_TEXTURE_2D)
                ourTexture = texture.Texture(
                    Image.open(os.path.join(HERE, 'yingyang.png'))
                )
                ourTexture()

                result = glGetTexImageub(GL_TEXTURE_2D, 0, GL_RGBA)
                assert isinstance(result, bytes), type(result)
                result = glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE)
                assert isinstance(result, bytes), type(result)

                glEnable(GL_LIGHTING)
                glEnable(GL_LIGHT0)
                glBegin(GL_TRIANGLES)
                try:
                    try:
                        glTexCoord2f(0.5, 1)
                        glVertex3f(0.0, 1.0, 0.0)
                    except Exception:
                        traceback.print_exc()
                    glTexCoord2f(0, 0)
                    glVertex3fv([-1, 0, 0])
                    glTexCoord2f(1, 0)
                    glVertex3dv([1, 0, 0])
                    try:
                        glVertex3dv([1, 0])
                    except ValueError:
                        assert (
                            OpenGL.ARRAY_SIZE_CHECKING
                        ), """Should have raised ValueError when doing array size checking"""
                    else:
                        assert (
                            not OpenGL.ARRAY_SIZE_CHECKING
                        ), """Should not have raised ValueError when not doing array size checking"""
                finally:
                    glEnd()

    @pytest.mark.skipif(not np, reason="Numpy not available")
    def test_numpyConversion(self):
        """Test that we can run a numpy conversion from double to float for glColorArray"""
        a = np.arange(0, 1.2, 0.1, 'd').reshape((-1, 3))
        glEnableClientState(GL_VERTEX_ARRAY)
        try:
            glColorPointerf(a)
            glColorPointerd(a)
        finally:
            glDisableClientState(GL_VERTEX_ARRAY)

    @pytest.mark.skipif(not np, reason="Numpy not available")
    def test_glbuffersubdata_numeric(self):
        from OpenGL.arrays import vbo

        assert vbo.get_implementation()
        points = np.array(
            [
                [0, 0, 0],
                [0, 1, 0],
                [1, 0.5, 0],
                [1, 0, 0],
                [1.5, 0.5, 0],
                [1.5, 0, 0],
            ],
            dtype='f',
        )
        d = vbo.VBO(points)
        with d:
            glBufferSubData(
                d.target,
                12,
                12,
                np.array([1, 1, 1], dtype='f'),
            )

    @pytest.mark.skipif(not np, reason="Numpy not available")
    @pytest.mark.skipif(OpenGL.ERROR_ON_COPY, reason="Test requires array copy")
    def test_copyNonContiguous(self):
        """Test that a non-contiguous (transposed) array gets applied as a copy"""
        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        try:
            transf = np.identity(4, dtype=np.float32)
            # some arbitrary transformation...
            transf[0, 3] = 2.5
            transf[2, 3] = -80

            # what do we get with the un-transposed version...
            glMatrixMode(GL_MODELVIEW)
            glLoadIdentity()
            glMultMatrixf(transf)
            untransposed = glGetFloatv(GL_MODELVIEW_MATRIX)
            # now transposed...

            # with a copy it works...
            t2 = transf.transpose().copy()
            # This doesn't work:
            glLoadIdentity()
            glMultMatrixf(t2)
            # This does work:
            # glMultMatrixf(transf.transpose().copy())
            transposed = glGetFloatv(GL_MODELVIEW_MATRIX)

            assert not np.allclose(transposed, untransposed), (transposed, untransposed)

            t2 = transf.transpose()
            # This doesn't work:
            glLoadIdentity()
            glMultMatrixf(t2)
            # This does work:
            # glMultMatrixf(transf.transpose().copy())
            transposed = glGetFloatv(GL_MODELVIEW_MATRIX)

            assert not np.allclose(transposed, untransposed), (transposed, untransposed)
        finally:
            glMatrixMode(GL_MODELVIEW)
            glPopMatrix()

    def test_bytes_array_support(self):
        color = b'\000' * 12
        glColor3fv(color)

    @pytest.mark.skipif(
        not (OpenGL.USE_ACCELERATE and acceleratesupport.ACCELERATE_AVAILABLE),
        reason="Need OpenGL_accelerate for buffer support",
    )
    def test_bytearray_support(self):
        import struct

        data = struct.pack(b'fff', 0.5, 0.4, 0.3)
        color = bytearray(data)
        glColor3fv(color)

    @pytest.mark.skipif(
        not (OpenGL.USE_ACCELERATE and acceleratesupport.ACCELERATE_AVAILABLE),
        reason="Need OpenGL_accelerate for buffer support",
    )
    def test_buffer_api_basic(self):
        import array as silly_array

        structures = []
        if sys.version_info[:2] >= (3, 9):
            structures.append(
                (b'this and that', 13, 1, True, 1, b'B', [13], [1]),
            )

        if sys.version_info[:2] >= (2, 7):
            # GH#92 Big-endian hosts report different formats because yeah, obviously
            if sys.byteorder == 'little':  # x86 cases
                int_formats = [b'(3)<i', b'(3)<l', b'<i', b'<l']
            else:
                int_formats = [b'(3)>i', b'(3)>l', b'>i', b'>l']

            if sys.version_info[:2] not in [(3, 8), (3, 7)]:

                structures.append(
                    # on Python 3.4 we do *not* get the (3) prefix :(
                    (
                        (GLint * 3)(1, 2, 3),
                        12,
                        4,
                        False,
                        1,
                        int_formats,
                        [3],
                        None,
                    ),
                )

        if sys.version_info[:2] >= (3, 0) and sys.version_info[:2] not in [
            (3, 8),
            (3, 7),
        ]:
            # only supports buffer protocol in 3.x
            structures.extend(
                [
                    (
                        silly_array.array('I', [1, 2, 3]),
                        12,
                        4,
                        False,
                        1,
                        b'I',
                        [3],
                        [4],
                    ),
                ]
            )
        try:
            if sys.version_info[:2] not in [(3, 8), (3, 7)]:
                structures.append((memoryview(b'this'), 4, 1, True, 1, b'B', [4], [1]))
        except NameError:
            # Python 2.6 doesn't have memory view
            pass
        try:
            if array:
                structures.extend(
                    [
                        (
                            arange(0, 9, dtype='I').reshape((3, 3)),
                            36,
                            4,
                            False,
                            2,
                            b'I',
                            [3, 3],
                            [12, 4],
                        ),
                        (
                            arange(0, 9, dtype='I').reshape((3, 3))[:, 1],
                            12,
                            4,
                            False,
                            1,
                            b'I',
                            [3],
                            [12],
                        ),
                    ]
                )
        except NameError:
            # Don't have numpy installed...
            pass

        from OpenGL.arrays import _buffers

        for (
            object,
            length,
            itemsize,
            readonly,
            ndim,
            format,
            shape,
            strides,
        ) in structures:
            buf = _buffers.Py_buffer.from_object(
                object, _buffers.PyBUF_STRIDES | _buffers.PyBUF_FORMAT
            )
            with buf:
                assert buf.len == length, (object, length, buf.len)
                assert buf.itemsize == itemsize, (object, itemsize, buf.itemsize)
                assert buf.readonly == readonly, (object, readonly, buf.readonly)
                assert buf.ndim == ndim, (object, ndim, buf.ndim)
                if isinstance(format, list):
                    assert buf.format in format, (object, format, buf.format)
                else:
                    assert buf.format == format, (object, format, buf.format)
                assert buf.shape[: buf.ndim] == shape, (
                    object,
                    shape,
                    buf.shape[: buf.ndim],
                )
                assert buf.dims == shape, (object, shape, buf.dims)
                assert buf.buf
                if strides is None:
                    assert not buf.strides
                else:
                    assert buf.strides[: buf.ndim] == strides, (
                        object,
                        strides,
                        buf.strides[: buf.ndim],
                    )
            assert buf.obj == None, buf.obj
            del buf

    @pytest.mark.skipif(
        not (OpenGL.USE_ACCELERATE and acceleratesupport.ACCELERATE_AVAILABLE),
        reason="Need OpenGL_accelerate for buffer support",
    )
    def test_memoryview_support(self):
        color = bytearray(b'\000' * 12)
        mem = memoryview(color)
        glColor3fv(mem)

    def test_void_dp_for_void_dp_is_self(self):
        array = ctypes.c_voidp(12)
        translated = ArrayDatatype.voidDataPointer(array)
        assert translated.value == array.value, translated

    def test_params_python3_strings(self):
        try:
            glGetUniformBlockIndex(0, unicode("Moo"))
        except ArgumentError:
            assert (
                OpenGL.ERROR_ON_COPY
            ), """Shouldn't have raised error on copy for unicode"""
        except TypeError:
            raise
        except GLError:
            # expected error, as we don't have a shader there...
            pass

    @pytest.mark.skipif(not np, reason="Numpy not available")
    def test_array_subclass(self):
        s = Subclassed([0, 1, 2, 3, 4])
        result = arraydatatype.ArrayDatatype.asArray(s)
        assert isinstance(result, Subclassed)

    @pytest.mark.skipif(not np, reason="Numpy not available")
    def test_byte_count_numpy(self):
        for a, expected in [
            (np.array([1, 2], dtype='e'), 4),
            (np.array([1, 2], dtype='f'), 8),
            (np.array([1, 2], dtype='d'), 16),
            (np.array([1, 2], dtype='B'), 2),
            (np.array([[1], [2]], dtype='B'), 2),
            (np.float32(), 4),
            (np.float64(), 8),
        ]:
            handler = arraydatatype.ArrayDatatype.getHandler(a)

            assert type(a) in handler.HANDLED_TYPES, type(a)
            calculated = arraydatatype.ArrayDatatype.arrayByteCount(a)
            assert calculated == expected, "Byte count for %s was %s, expected %" % (
                a,
                calculated,
                expected,
            )

    def test_byte_count(self):
        for a, expected in [
            (ctypes.c_float(), 4),
            ((ctypes.c_float * 3 * 4)(), 4 * 3 * 4),
        ]:
            calculated = arraydatatype.ArrayDatatype.arrayByteCount(a)
            assert calculated == expected, "Byte count for %s was %s, expected %" % (
                a,
                calculated,
                expected,
            )


if np:

    class Subclassed(np.ndarray):
        pass



#! /usr/bin/env python3
"""Raw xlib based test setup"""
from Xlib import X, display, error
import ctypes
from functools import wraps

from OpenGL import arrays
from OpenGL.EGL import *

DESIRED_ATTRIBUTES = [
    EGL_BLUE_SIZE,
    8,
    EGL_RED_SIZE,
    8,
    EGL_GREEN_SIZE,
    8,
    EGL_DEPTH_SIZE,
    24,
    EGL_COLOR_BUFFER_TYPE,
    EGL_RGB_BUFFER,
    EGL_CONFIG_CAVEAT,
    EGL_NONE,  # Don't allow slow/non-conformant
]
API_BITS = {
    "opengl": EGL_OPENGL_BIT,
    "gl": EGL_OPENGL_BIT,
    "gles2": EGL_OPENGL_ES2_BIT,
    "gles1": EGL_OPENGL_ES_BIT,
    "gles": EGL_OPENGL_ES_BIT,
    "es2": EGL_OPENGL_ES2_BIT,
    "es1": EGL_OPENGL_ES_BIT,
    "es": EGL_OPENGL_ES_BIT,
}
API_NAMES = dict(
    [
        (
            k,
            {
                EGL_OPENGL_BIT: EGL_OPENGL_API,
                EGL_OPENGL_ES2_BIT: EGL_OPENGL_ES_API,
                EGL_OPENGL_ES_BIT: EGL_OPENGL_ES_API,
            }[v],
        )
        for k, v in API_BITS.items()
    ]
)


class EGLWindow(object):
    def __init__(
        self, display, msg, size=(300, 300), api='es2', attributes=DESIRED_ATTRIBUTES
    ):
        self.display = display
        self.msg = msg

        self.screen: Xlib.Window = self.display.screen()
        event_mask = (
            X.ExposureMask
            | X.ResizeRedirectMask
            | X.StructureNotifyMask
            | X.ButtonPressMask
            | X.ButtonReleaseMask
            | X.Button1MotionMask
        )
        self.window = self.screen.root.create_window(
            50,
            50,
            size[0],
            size[1],
            2,
            self.screen.root_depth,
            X.InputOutput,
            X.CopyFromParent,
            # special attribute values
            background_pixel=self.screen.white_pixel,
            event_mask=event_mask,
            colormap=X.CopyFromParent,
        )
        self.gc = self.window.create_gc(
            foreground=self.screen.black_pixel,
            background=self.screen.white_pixel,
        )
        self.window.map()
        self.api = api
        self.attributes = attributes

    egl_ctx = None

    def eglSetup(self):
        """Setup the EGL rendering context"""
        if self.egl_ctx is not None:
            return
        major, minor = ctypes.c_long(), ctypes.c_long()
        self.egl_display = display = eglGetDisplay(EGL_DEFAULT_DISPLAY)
        eglInitialize(display, major, minor)
        num_configs = ctypes.c_long()
        configs = (EGLConfig * 2)()
        api_constant = API_NAMES[self.api.lower()]
        print('api requested:', self.api, ' api constant', api_constant)
        local_attributes = self.attributes[:]
        local_attributes.extend(
            [
                EGL_CONFORMANT,
                API_BITS[self.api.lower()],
                EGL_NONE,
            ]
        )
        print("local_attributes", local_attributes)
        local_attributes = arrays.GLintArray.asArray(local_attributes)
        eglChooseConfig(display, local_attributes, configs, 2, num_configs)
        if num_configs.value < 1:
            raise RuntimeError("Unable to find a suitable config")
        print("API", api_constant)
        eglBindAPI(api_constant)
        window = self.window.id
        self.egl_surface = eglCreateWindowSurface(display, configs[0], window, None)

        self.egl_ctx = eglCreateContext(display, configs[0], EGL_NO_CONTEXT, None)
        if self.egl_ctx == EGL_NO_CONTEXT:
            raise RuntimeError("Unable to create context")

    def loop(self, target, args, named, exit_on_render=False):
        do_close = True
        try:
            while True:
                try:
                    e = self.display.next_event()
                except error.ConnectionClosedError as err:
                    do_close = False
                    return

                if e.type == X.Expose:
                    self.eglSetup()
                    eglMakeCurrent(
                        self.egl_display,
                        self.egl_surface,
                        self.egl_surface,
                        self.egl_ctx,
                    )
                    target(*args, **named)
                    eglSwapBuffers(self.egl_display, self.egl_surface)
                    if exit_on_render:
                        return

                elif e.type == X.KeyPress:
                    return
        finally:
            if do_close:
                self.close()

    def close(self):
        self.window.destroy()


def egltest(size=(300, 300), name=None, api="es2", attributes=DESIRED_ATTRIBUTES):
    def gltest(function):
        """Decorator to allow a function to run in a Pygame GLES[1,2,3] context"""

        @wraps(function)
        def test_function(*args, **named):
            window = EGLWindow(
                display.Display(),
                name or function.__name__,
                api=api,
                attributes=attributes,
            )
            window.loop(target=function, args=args, named=named, exit_on_render=True)

        return test_function

    return gltest


if __name__ == "__main__":

    @egltest()
    def test_sample(*args, **named):
        print("Ran sample")

    test_sample()
    test_sample()
    test_sample()



from os_egl import egl_context
from OpenGL import EGL
from OpenGL.EGL.EXT import device_query, device_enumeration
from OpenGL.GL import GLint


def main():
    with egl_context(output=None, pbuffer=True) as context:
        display, context, surface = context
        print("Vendor: %s" % (EGL.eglQueryString(display, EGL.EGL_VENDOR)))
        print("Version: %s" % (EGL.eglQueryString(display, EGL.EGL_VERSION)))
        print("Extensions: %s" % (EGL.eglQueryString(display, EGL.EGL_EXTENSIONS),))
        print(
            "Client Extensions: %s"
            % (EGL.eglQueryString(display, EGL.EGL_CLIENT_APIS),)
        )
        if device_enumeration.eglQueryDevicesEXT:
            devices = (device_query.EGLDeviceEXT * 5)()
            count = GLint()
            device_enumeration.eglQueryDevicesEXT(
                5,
                devices,
                count,
            )
            for device in devices[: count.value]:
                print(device)
        else:
            print('No device_query extension available')
        print('OK')


if __name__ == "__main__":
    import logging

    logging.basicConfig(level=logging.DEBUG)
    main()



#! /usr/bin/env python
import xlibegltest as egltest
import numpy
from OpenGL import GL


@egltest.egltest(api='opengl')
def test_gl():
    GL.glClearColor(1, 0, 0, 0)
    GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
    print('OK')


if __name__ == "__main__":
    test_gl()



"""Transliteration of https://open.gl/feedback into Python"""
from __future__ import print_function
import pygamegltest
from OpenGL._bytes import as_8_bit
from OpenGL.GL import *
from OpenGL.GL import shaders

vertex_shader = """#version 150 core
    in float inValue;
    out float geoValue;
    uniform float junk; // Just here for the introspection test

    void main() {
        geoValue = sqrt(inValue) + junk;
    }
"""

geometry_shader = """#version 150 core
    layout(points) in;
    layout(triangle_strip, max_vertices = 3) out;

    in float[] geoValue;
    out float outValue;

    void main() {
        for (int i = 0; i < 3; i++) {
            outValue = geoValue[0] + i;
            EmitVertex();
        }

        EndPrimitive();
    }
"""


@pygamegltest.pygametest(name="Geometry Shader Test")
def main():
    program = shaders.compileProgram(
        shaders.compileShader([vertex_shader], GL_VERTEX_SHADER),
        shaders.compileShader([geometry_shader], GL_GEOMETRY_SHADER),
    )

    # test that we can describe the attributes in the shader
    for i in range(glGetProgramiv(program, GL_ACTIVE_ATTRIBUTES)):
        name, size, type = glGetActiveAttrib(program, i)
        print('Atribute %s(%i) -> %s %s' % (name, i, size, type))

        length, size, type, name = GLsizei(), GLint(), GLenum(), (GLchar * 8)()
        glGetActiveAttrib(program, i, None, length, size, type, name)
        assert length.value == len(name.value), (length.value, name.value)
        assert type.value == 5126, type.value

    for i in range(glGetProgramiv(program, GL_ACTIVE_UNIFORMS)):
        name, size, type = glGetActiveUniform(program, i)
        print('Uniform %s(%i) -> %s %s' % (name, i, size, type))

        length, size, type, name = GLsizei(), GLint(), GLenum(), (GLchar * 5)()
        glGetActiveUniform(program, i, 5, length, size, type, name)
        assert length.value == len(name.value), (length.value, name.value)
        assert type.value == 5126, type.value

    buff = (ctypes.c_char_p * 1)(as_8_bit("outValue\000"))
    c_text = ctypes.cast(ctypes.pointer(buff), ctypes.POINTER(ctypes.POINTER(GLchar)))
    # modifies the state in the linking of the program
    glTransformFeedbackVaryings(program, 1, c_text, GL_INTERLEAVED_ATTRIBS)
    # so have to re-link
    glLinkProgram(program)

    glUseProgram(program)

    vao = glGenVertexArrays(1)
    glBindVertexArray(vao)

    data = (GLfloat * 5)(1.0, 2.0, 3.0, 4.0, 5.0)

    vbo = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, vbo)
    glBufferData(GL_ARRAY_BUFFER, ctypes.sizeof(data), data, GL_STATIC_DRAW)

    inputAttrib = glGetAttribLocation(program, "inValue")
    glEnableVertexAttribArray(inputAttrib)
    glVertexAttribPointer(inputAttrib, 1, GL_FLOAT, GL_FALSE, 0, 0)

    tbo = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, tbo)
    glBufferData(GL_ARRAY_BUFFER, ctypes.sizeof(data) * 3, None, GL_STATIC_READ)

    glEnable(GL_RASTERIZER_DISCARD)

    glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, 0, tbo)

    glBeginTransformFeedback(GL_TRIANGLES)
    glDrawArrays(GL_POINTS, 0, 5)
    glEndTransformFeedback()

    glDisable(GL_RASTERIZER_DISCARD)

    glFlush()

    feedback = (GLfloat * 15)(*([0] * 15))
    glGetBufferSubData(
        GL_TRANSFORM_FEEDBACK_BUFFER, 0, ctypes.sizeof(feedback), feedback
    )

    for value in feedback:
        print(value)


if __name__ == "__main__":
    main()



#! /usr/bin/env python
"""Script to find missing GLUT entry points"""
from OpenGL import GLUT 
import subprocess, re
func_finder = re.compile( 'FGAPIENTRY (\w+)\(' )
constant_finder = re.compile( '#define\W+([0-9a-zA-Z_]+)\W+((0x)?\d+)' )

INCLUDE_DIR = '/usr/include/GL'

def defined( ):
    """Grep FGAPIENTRY headers from /usr/include/GL"""
    pipe = subprocess.Popen( 'grep -r FGAPIENTRY %(INCLUDE_DIR)s/*'%globals(), shell=True, stdout=subprocess.PIPE )
    stdout,stderr = pipe.communicate()
    return stdout
def constants():
    pipe = subprocess.Popen( 'grep -r "#define" %(INCLUDE_DIR)s/*glut*'%globals(), shell=True, stdout=subprocess.PIPE )
    stdout,stderr = pipe.communicate()
    return stdout

def main():
    headers = {}
    for line in defined().splitlines():
        match = func_finder.search( line )
        if match:
            headers[match.group(1)] = line.split(':',1)[0]
    for key in headers.keys():
        if hasattr( GLUT, key ):
            del headers[key]
    import pprint 
    pprint.pprint( headers )
    missing = {}
    for line in constants().splitlines():
        match = constant_finder.search( line )
        if match:
            key,value=(match.group(1),match.group(2))
            if not hasattr( GLUT, key ):
                file = line.split(':',1)[0]
                missing.setdefault(file,[]).append( (key,value))
    for file,variables in missing.items():
        print file
        variables.sort()
        for key,value in variables:
            print '%s=%s'%(key,value)
        

if __name__ == "__main__":
    main()



from __future__ import print_function
from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.GLU import *
import time, sys

resX, resY = (400, 300)


def display():
    glutSetWindow(window)
    glClearColor(0.0, 0.0, (time.time() % 1.0) / 1.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT)
    glFlush()
    glutSwapBuffers()
    sys.stdout.write('OK\n')
    sys.stdout.flush()
    if glutLeaveMainLoop:
        glutLeaveMainLoop()


if __name__ == "__main__":
    glutInit([])
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH)

    glutInitWindowSize(resX, resY)

    glutInitWindowPosition(0, 0)
    window = glutCreateWindow("hello")
    glutDisplayFunc(display)
    for name in (GL_VENDOR, GL_RENDERER, GL_SHADING_LANGUAGE_VERSION, GL_EXTENSIONS):
        print(('%s = %r' % (name, glGetString(name))))

    glutMainLoop()



import logging, os, urllib, traceback, textwrap, keyword
import xmlreg
from OpenGL._bytes import as_8_bit, as_str, unicode, as_unicode

HERE = os.path.join(os.path.dirname(__file__))
log = logging.getLogger(__name__)
AUTOGENERATION_SENTINEL = (
    """### DO NOT EDIT above the line "END AUTOGENERATED SECTION" below!"""
)
AUTOGENERATION_SENTINEL_END = """### END AUTOGENERATED SECTION"""


def nameToPathMinusGL(name):
    return "/".join(name.split('_', 2)[1:])


def prefix_split(name):
    return name.split('_', 2)[1:]


def indent(text, indent='\t'):
    return "\n".join(['%s%s' % (indent, line) for line in text.splitlines()])


class Generator(object):
    targetDirectory = os.path.join(HERE, '..', 'OpenGL')
    rawTargetDirectory = os.path.join(HERE, '..', 'OpenGL', 'raw')
    prefix = 'GL'
    dll = '_p.PLATFORM.GL'
    includeOverviews = True

    def __init__(self, registry, type_translator):
        self.registry = registry
        self.type_translator = type_translator

    def module(self, module):
        if hasattr(module, 'apis'):
            for api in module.apis:
                if api != 'glcore':
                    gen = ModuleGenerator(module, self, api)
                    gen.generate()
        else:
            gen = ModuleGenerator(module, self)
            gen.generate()
        return gen

    GLGET_PARAM_GROUPS = [
        #'MaterialParameter',
        #'PixelMap',
        #'LightParameter',
        'GetPName',
        #'GetPixelMap',
        #'GetMapQuery',
        'GetPointervPName',
        #'TextureEnvParameter',
        #'TextureGenParameter',
    ]
    GL_GET_TEMPLATE = '''"""glGet* auto-generation of output arrays (DO NOT EDIT, AUTOGENERATED)"""
try:
    from OpenGL.raw.%(prefix)s._lookupint import LookupInt as _L
except ImportError:
    def _L(*args):
        raise RuntimeError( "Need to define a lookupint for this api" )
_glget_size_mapping = _m = {}
%(elements)s
'''

    def group_sizes(self):
        """Generate a group-sizes data-table for the given group-name"""
        result = []
        for enum_name, size in sorted(
            self.glGetSizes.items(), key=lambda x: (bool(x[1]), x[0])
        ):
            value = self.registry.enumeration_set.get(enum_name)
            size = [x for x in size if x]
            comment = ''
            if not size:
                size = 'TODO'
                comment = '# '
            else:
                size = ''.join(size)
            if value is None:
                # common in cases where GL and GLES constants are updated together...
                log.debug('Unrecognized constant: %s in GLGet section', enum_name)
            else:
                value = value.value
                result.append(
                    '%(comment)s_m[%(value)s] = %(size)s # %(enum_name)s' % locals()
                )
        elements = '\n'.join(result)
        prefix = self.prefix
        return self.GL_GET_TEMPLATE % locals()

    def enum(self, enum):
        comment = ''
        try:
            int(enum.value, 0)
        except ValueError:
            comment = '# '
        return '%s%s=_C(%r,%s)' % (comment, enum.name, enum.name, enum.value)

    def safe_name(self, name):
        if keyword.iskeyword(name):
            return name + '_'
        return name

    def function(self, function):
        """Produce a declaration for this function in ctypes format"""
        returnType = self.type_translator(function.returnType)
        if returnType == 'arrays.GLbyteArray':
            returnType = 'ctypes.c_char_p'
        if function.argTypes:
            argTypes = ','.join([self.type_translator(x) for x in function.argTypes])
        else:
            argTypes = ''
        if function.argNames:
            argNames = ','.join([self.safe_name(n) for n in function.argNames])
        else:
            argNames = ''
        arguments = ', '.join(
            [
                '%s %s' % (t, self.safe_name(n))
                for (t, n) in zip(function.argTypes, function.argNames)
            ]
        )
        name = function.name
        if returnType.strip() in ('_cs.GLvoid', '_cs.void', 'void'):
            returnType = pyReturn = 'None'
        else:
            pyReturn = function.returnType
        doc = '%(name)s(%(arguments)s) -> %(pyReturn)s' % locals()
        #        log.info( '%s', doc )
        formatted = self.FUNCTION_TEMPLATE % locals()
        return formatted

    FUNCTION_TEMPLATE = """@_f
@_p.types(%(returnType)s,%(argTypes)s)
def %(name)s(%(argNames)s):pass"""

    _glGetSizes = None

    @property
    def glGetSizes(self):
        if self._glGetSizes is None:
            self._glGetSizes = self.loadGLGetSizes()
        return self._glGetSizes

    def loadGLGetSizes(self):
        """Load manually-generated table of glGet* sizes"""
        table = {}
        try:
            lines = [
                line.split('\t')
                for line in open(os.path.join(HERE, 'glgetsizes.csv'))
                .read()
                .splitlines()
            ]
        except IOError:
            pass
        else:
            for line in lines:
                if line and line[0]:
                    value = [v for v in [v.strip('"') for v in line[1:]] if v]
                    if value:
                        table[line[0].strip('"').strip()] = value
        # now make sure everything registered in the xml file is present...
        output_group_names = {}
        for function in self.registry.command_set.values():
            output_group_names.update(function.output_groups)
        for output_group in output_group_names.keys():
            log.debug('Output parameter group: %s', output_group)
            for name in self.registry.enum_groups.get(output_group, []):
                if name not in table:
                    log.info('New %s value: %r', output_group, name)
                    table[name] = ''
        return table

    def saveGLGetSizes(self):
        """Save out sorted list of glGet sizes to disk"""
        items = list(self.glGetSizes.items())
        items.sort()
        data = "\n".join(['%s\t%s' % (key, "\t".join(value)) for (key, value) in items])
        open(os.path.join(HERE, 'glgetsizes.csv'), 'w').write(data)


class ModuleGenerator(object):
    ROOT_EXTENSION_SOURCE = 'http://www.opengl.org/registry/specs/'
    RAW_MODULE_TEMPLATE = """'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.%(prefix)s import _types as _cs
# End users want this...
from OpenGL.raw.%(prefix)s._types import *
from OpenGL.raw.%(prefix)s import _errors
from OpenGL.constant import Constant as _C
%(extra_imports)s
import ctypes
_EXTENSION_NAME = %(constantModule)r
def _f( function ):
    return _p.createFunction( function,%(dll)s,%(constantModule)r,error_checker=_errors._error_checker)
%(constants)s
%(declarations)s
"""

    INIT_TEMPLATE = """
def glInit%(camelModule)s%(owner)s():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )
"""
    FINAL_MODULE_TEMPLATE = """'''OpenGL extension %(owner)s.%(module)s

This module customises the behaviour of the 
OpenGL.raw.%(prefix)s.%(owner)s.%(module)s to provide a more 
Python-friendly API

%(overview)sThe official definition of this extension is available here:
%(ROOT_EXTENSION_SOURCE)s%(owner)s/%(module)s.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.%(prefix)s import _types, _glgets
from OpenGL.raw.%(prefix)s.%(owner)s.%(module)s import *
from OpenGL.raw.%(prefix)s.%(owner)s.%(module)s import _EXTENSION_NAME
%(init_function)s
%(output_wrapping)s
"""
    dll = '_p.PLATFORM.GL'

    def __init__(self, registry, overall, api=None):
        self.registry = registry
        self.overall = overall
        name = registry.name
        if name in ('GL_ES_VERSION_3_1', 'GL_ES_VERSION_3_0'):
            api = 'gles3'
            name = 'GLES3' + name[5:]
        if api:
            self.prefix = api.upper()
        else:
            if hasattr(self.registry, 'api'):
                self.prefix = self.registry.api.upper()
            else:
                self.prefix = name.split('_')[0]
        name = name.split('_', 1)[1]
        try:
            self.owner, self.module = name.split('_', 1)
            self.sentinelConstant = '%s_%s' % (self.owner, self.module)

        except ValueError:
            if name.endswith('SGIX'):
                self.prefix = "GL"
                self.owner = 'SGIX'
                self.module = name[3:-4]
                self.sentinelConstant = '%s%s' % (self.module, self.owner)
            else:
                log.error("""Unable to parse module name: %s""", name)
                raise
        self.dll = '_p.PLATFORM.%s' % (self.prefix,)
        if self.module[0].isdigit():
            self.module = '%s_%s' % (
                self.prefix,
                self.module,
            )
        if self.module == 'async':
            self.module = 'async_'
        self.camelModule = "".join([x.title() for x in self.module.split('_')])
        self.rawModule = self.module

        self.rawOwner = self.owner
        while self.owner and self.owner[0].isdigit():
            self.owner = self.owner[1:]
        self.rawPathName = os.path.join(
            self.overall.rawTargetDirectory,
            self.prefix,
            self.owner,
            self.module + '.py',
        )
        self.pathName = os.path.join(
            self.overall.targetDirectory, self.prefix, self.owner, self.module + '.py'
        )

        self.constantModule = '%(prefix)s_%(owner)s_%(rawModule)s' % self
        specification = self.getSpecification()
        self.constantsFromSpec()
        self.overview = ''
        if self.overall.includeOverviews:
            for title, section in specification.blocks(specification.source):
                if title.startswith('Overview'):
                    if isinstance(section, bytes):
                        section = section.decode(
                            'latin-1'
                        )  # seems not to be utf in at lest some cases
                    self.overview = 'Overview (from the spec)\n%s\n\n' % (
                        indent(
                            as_unicode(
                                as_str(section)
                                .replace('\xd4', 'O')
                                .replace('\xd5', 'O')
                            )
                            .encode('ascii', 'ignore')
                            .decode('ascii', 'ignore')
                            .replace('\\', '\\\\')
                        )
                    )
                    break

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __getattr__(self, key):
        if key not in ('registry',):
            return getattr(self.registry, key)

    @property
    def extra_imports(self):
        if self.name == 'GL_VERSION_1_1':
            # spec files have not properly separated out these two...
            return '# Spec mixes constants from 1.0 and 1.1\nfrom OpenGL.raw.GL.VERSION.GL_1_0 import *'
        return ''

    def constantsFromSpec(self, spec=None):
        """Examine spec text looking for new constants..."""
        spec = spec or self.getSpecification()
        if spec.source:
            table = self.overall.glGetSizes
            extras = spec.glGetConstants()
            changed = False
            for key, value in extras.items():
                if key not in table:
                    try:
                        value = int(value, 16)
                    except ValueError:
                        continue
                    short_key = '_'.join(key.split("_")[:-1])
                    if short_key in table:
                        short_def = self.overall.registry.enumeration_set.get(short_key)
                        if short_def and int(short_def.value, 16) == value:
                            key = short_def.name
                    if not key in table:
                        table[key] = ['(1,)', '#TODO Review %s' % (self.specFile())]
                        changed = True
            if changed:
                self.overall.saveGLGetSizes()

    def shouldReplace(self):
        """Should we replace the given filename?"""
        filename = self.pathName
        if not os.path.isfile(filename):
            return True
        else:
            hasLines = 0
            for line in open(filename):
                if line.strip() == AUTOGENERATION_SENTINEL_END.strip():
                    return True
                hasLines = 1
            if not hasLines:
                return True
            log.warning(
                'Not replacing %s (no AUTOGENERATION_SENTINEL_END found)', filename
            )
        return False

    @property
    def output_wrapping(self):
        """Generate output wrapping statements for our various functions"""
        try:
            statements = []
            for function in self.registry.commands():
                dependencies = function.size_dependencies
                if dependencies:  # temporarily just do single-output functions...
                    base = []
                    for param, dependency in sorted(dependencies.items()):
                        param = as_str(param)
                        if isinstance(dependency, xmlreg.Output):
                            statements.append(
                                '# %s.%s is OUTPUT without known output size'
                                % (
                                    function.name,
                                    param,
                                )
                            )
                        if isinstance(dependency, xmlreg.Staticsize):
                            base.append(
                                '.setOutput(\n    %(param)r,size=(%(dependency)r,),orPassIn=True\n)'
                                % locals()
                            )
                        elif isinstance(dependency, xmlreg.Dynamicsize):
                            if '/' in dependency:
                                fragments = [x.strip() for x in dependency.split('/')]
                                functor = 'lambda x: (x//%s)' % (fragments[1])
                                dependency = fragments[0]
                            elif '*' in dependency:
                                fragments = [x.strip() for x in dependency.split('/')]
                                functor = 'lambda x: (x*%s)' % (fragments[1])
                                dependency = fragments[0]
                            else:
                                functor = 'lambda x:(x,)'

                            base.append(
                                '.setOutput(\n    %(param)r,size=%(functor)s,pnameArg=%(dependency)r,orPassIn=True\n)'
                                % locals()
                            )
                        elif isinstance(dependency, xmlreg.Multiple):
                            pname, multiple = dependency
                            base.append(
                                '.setOutput(\n    %(param)r,size=lambda x:(x,%(multiple)s),pnameArg=%(pname)r,orPassIn=True\n)'
                                % locals()
                            )
                        elif isinstance(dependency, xmlreg.Compsize):
                            if len(dependency) == 1:
                                pname = dependency[0]
                                base.append(
                                    '.setOutput(\n    %(param)r,size=_glgets._glget_size_mapping,pnameArg=%(pname)r,orPassIn=True\n)'
                                    % locals()
                                )
                            else:
                                statements.append(
                                    '# OUTPUT %s.%s COMPSIZE(%s) '
                                    % (function.name, param, ', '.join(dependency))
                                )
                        elif isinstance(dependency, xmlreg.StaticInput):
                            base.append(
                                '.setInputArraySize(\n    %(param)r, %(dependency)s\n)'
                                % locals()
                            )
                        elif isinstance(
                            dependency,
                            (xmlreg.DynamicInput, xmlreg.MultiplyInput, xmlreg.Input),
                        ):
                            if dependency is None:
                                continue
                            statements.append(
                                '# INPUT %s.%s size not checked against %s'
                                % (function.name, param, dependency)
                            )
                            base.append(
                                '.setInputArraySize(\n    %(param)r, None\n)' % locals()
                            )
                    if base:
                        base.insert(
                            0, '%s=wrapper.wrapper(%s)' % (function.name, function.name)
                        )
                        statements.append(''.join(base))
            return '\n'.join(statements)
        except Exception as err:
            traceback.print_exc()
            import pdb

            pdb.set_trace()

    def get_constants(self):
        functions = self.registry.enums()
        functions.sort(key=lambda x: x.name)
        return functions

    @property
    def init_function(self):
        return self.INIT_TEMPLATE % self

    @property
    def constants(self):
        result = []
        try:
            for function in self.get_constants():
                result.append(self.overall.enum(function))
            return '\n'.join(result)
        except Exception:
            traceback.print_exc()
            raise

    @property
    def declarations(self):
        functions = self.registry.commands()
        functions.sort(key=lambda x: x.name)
        result = []
        for function in functions:
            result.append(self.overall.function(function))
        return "\n".join(result)

    EXTENSION_BASE = os.path.join(HERE, 'khronosapi', 'extensions')

    def specFile(self):
        """Lookup the specification file for our extension"""
        prefix, name = prefix_split(self.name)
        specFile = os.path.join(
            self.EXTENSION_BASE,
            prefix,
            '%s_%s.txt'
            % (
                prefix,
                name,
            ),
        )
        return specFile

    def getSpecification(self):
        """Retrieve our specification document...

        Retrieves the .txt file which defines this specification,
        allowing us to review the document locally in order to provide
        a reasonable wrapping of it...
        """
        if self.registry.feature:
            return Specification('')
        specFile = self.specFile()
        if os.path.exists(specFile):
            data = open(specFile, 'rb').read()
            return Specification(data)
        else:
            log.info("No spec-file in %s", specFile)
            return Specification('')

    def generate(self):
        for target in (self.rawPathName, self.pathName):
            directory = os.path.dirname(target)
            if not os.path.exists(directory):
                log.warning('Creating target directory: %s', directory)
                os.makedirs(directory)
            if not os.path.isfile(os.path.join(directory, '__init__.py')):
                open(os.path.join(directory, '__init__.py'), 'w').write(
                    '''"""OpenGL Extensions"""'''
                )

        directory = os.path.dirname(self.rawPathName)
        current = ''
        toWrite = self.RAW_MODULE_TEMPLATE % self
        try:
            current = open(self.rawPathName, 'r').read()
        except Exception as err:
            pass
        if current.strip() != toWrite.strip():
            fh = open(self.rawPathName, 'w')
            fh.write(toWrite)
            fh.close()
        if isinstance(self.registry, xmlreg.Feature):
            # this is a core feature...
            target = os.path.join(
                self.overall.rawTargetDirectory, self.prefix, '_glgets.py'
            )
            open(target, 'w').write(self.overall.group_sizes())
        if self.shouldReplace():
            # now the final module with any included custom code...
            toWrite = self.FINAL_MODULE_TEMPLATE % self
            current = ''
            try:
                current = open(self.pathName, 'r').read()
            except Exception as err:
                pass
            else:
                found = current.rfind('\n' + AUTOGENERATION_SENTINEL_END)
                if found >= -1:
                    if current[:found].strip() == toWrite.strip():
                        # we aren't going to change anything...
                        return False
                    found += len('\n' + AUTOGENERATION_SENTINEL_END)
                    current = current[found:]
                else:
                    current = ''
            try:
                fh = open(self.pathName, 'w')
            except IOError as err:
                log.warning("Unable to create module for %r %s", self.name, err)
                return False
            else:
                fh.write(toWrite)
                fh.write(AUTOGENERATION_SENTINEL_END)
                fh.write(current)
                fh.close()
                return True
        return False


class Specification(object):
    """Parser for parsing OpenGL specifications for interesting information"""

    def __init__(self, source):
        """Store the source text for the specification"""
        if isinstance(source, bytes):
            try:
                self.source = source.decode('utf-8')
            except UnicodeDecodeError as err:
                self.source = source.decode('latin-1')
        else:
            self.source = source
        assert isinstance(self.source, unicode)

    def blocks(self, data):
        """Retrieve the set of all blocks"""
        data = data.splitlines()
        title = []
        block = []
        for line in data:
            if line and line.lstrip() == line:
                if block:
                    yield "\n".join(title), textwrap.dedent("\n".join(block))
                    title = []
                    block = []
                title.append(line)
            else:
                block.append(line)
        if block:
            yield "\n".join(title), textwrap.dedent("\n".join(block))

    def constantBlocks(self):
        """Retrieve the set of constant blocks"""
        for title, block in self.blocks(self.source):
            if title and title.startswith('New Tokens'):
                yield block

    def glGetConstants(self):
        """Retrieve the set of constants which pass to glGet* functions"""
        table = {}
        for block in self.constantBlocks():
            for title, section in self.blocks(block):
                for possible in (
                    'GetBooleanv',
                    'GetIntegerv',
                    '<pname> of Get',
                    '<pname> parameter of Get',
                ):
                    if possible in title:
                        for line in section.splitlines():
                            line = line.strip().split()
                            if len(line) == 2:
                                constant, value = line
                                constant = constant.strip(':')
                                table['GL_%s' % (constant,)] = value
                        break
        return table


def download(url):
    """Download the given url, informing the user of what we're doing"""
    log.info(
        'Download: %r',
        url,
    )
    file = urllib.urlopen(url)
    return file.read()



from __future__ import print_function
import pytest, os, sys, unittest
import logging
log = logging.getLogger(__name__)
HERE = os.path.dirname( __file__ )
if sys.platform != 'win32':
    raise pytest.skip('Non-windows (WGL) platform', allow_module_level=True)
try:
    from numpy import *
except ImportError as err:
    array = None
import pygame, pygame.display
pygame.display.init()
from OpenGL.GL import *
from OpenGL.WGL import *

class TestWGL(unittest.TestCase):
    width,height = 300,300
    def setUp(self):
        self.screen = pygame.display.set_mode(
            (self.width,self.height),
            pygame.OPENGL | pygame.DOUBLEBUF,
        )
        
        pygame.display.set_caption('Testing system')
        pygame.key.set_repeat(500,30)
    def tearDown(self):
        pygame.display.flip()

    def test_wgl_imported(self):
        assert bool(wglCreateContext)
    def test_create_context(self):
        window = pygame.display.get_wm_info()['window']
        wglCreateContext(window)

    def test_get_extensions_low_level(self):    
        from OpenGL.WGL.ARB.extensions_string import wglGetExtensionsStringARB
        window = pygame.display.get_wm_info()['window']
        extensions = wglGetExtensionsStringARB(wglGetCurrentDC())
        assert extensions 
        assert b'WGL_ARB_extensions_string' in extensions, extensions

    def test_swap_control_interval(self):
        from OpenGL.WGL.EXT import swap_control
        if not swap_control.wglGetSwapIntervalEXT:
            raise pytest.skip(reason='No wglGetSwapIntervalEXT available')
        interval = swap_control.wglGetSwapIntervalEXT()

        swap_control.wglSwapIntervalEXT(1)



"""Utility function to convert a C type to a PyOpenGL-specific data-type name"""
import logging 
log = logging.getLogger( __name__ )

CTYPE_TO_ARRAY_TYPE = {
    '_cs.GLfloat': 'GLfloatArray',
    '_cs.float': 'GLfloatArray',
    'ctypes.c_float':'GLfloatArray',
    '_cs.GLclampf': 'GLclampfArray',
    '_cs.GLdouble': 'GLdoubleArray',
    '_cs.double': 'GLdoubleArray',
    '_cs.int': 'GLintArray',
    '_cs.GLint': 'GLintArray',
    '_cs.GLuint': 'GLuintArray',
    '_cs.unsigned int':'GLuintArray',
    '_cs.unsigned char': 'GLbyteArray',
    '_cs.uint': 'GLuintArray',
    '_cs.GLshort': 'GLshortArray',
    '_cs.GLushort': 'GLushortArray',
    '_cs.short unsigned int':'GLushortArray',
    '_cs.GLubyte': 'GLubyteArray',
    '_cs.GLbool': 'GLbooleanArray',
    '_cs.GLboolean': 'GLbooleanArray',
    'arrays.GLbooleanArray': 'GLbooleanArray',
    '_cs.GLbyte': 'GLbyteArray',
    '_cs.char': 'GLbyteArray',
    '_cs.gleDouble': 'GLdoubleArray',
    '_cs.GLchar': 'GLcharArray',
    '_cs.GLcharARB': 'GLcharARBArray',
    '_cs.GLhalfNV': 'GLushortArray',
    '_cs.GLhandle': 'GLuintArray',
    '_cs.GLhandleARB': 'GLuintArray',
    '_cs.GLenum': 'GLuintArray',
    # following should all have special sub-classes that enforce dimensions
    '_cs.gleDouble * 4': 'GLdoubleArray',
    '_cs.gleDouble * 3': 'GLdoubleArray',
    '_cs.gleDouble * 2': 'GLdoubleArray',
    '_cs.c_float * 3': 'GLfloatArray',
    '_cs.gleDouble * 3 * 2': 'GLdoubleArray',
    '_cs.GLsizei': 'GLsizeiArray',
    '_cs.GLint64': 'GLint64Array',
    '_cs.GLint64EXT': 'GLint64Array',
    '_cs.GLuint64': 'GLuint64Array',
    '_cs.GLuint64EXT': 'GLuint64Array',
    
    '_cs.EGLint':'GLintArray',
    '_cs.EGLConfig':'GLvoidpArray',
    '_cs.EGLuint64KHR':'GLuint64Array',
    '_cs.EGLNativeDisplayType':'GLvoidpArray',
    '_cs.EGLNativeWindowType': 'GLvoidpArray',
    '_cs.EGLNativePixmapType': 'GLvoidpArray',
    '_cs.EGLTimeKHR': 'GLuint64Array',
    
    '_cs.GLfixed':'GLfixedArray', 
    '_cs.EGLAttrib':'EGLAttribArray', 
}
SPECIAL_TYPES = {
    '__eglMustCastToProperFunctionPointerType': 'ctypes.c_void_p',
}


def ctype_to_pytype( base ):
    """Given a C declared type for an argument/return type, get Python/ctypes version"""
    base = base.strip()
    for strip in ('const','struct'):
        if base.endswith( strip ):
            return ctype_to_pytype( base[:-len(strip)] )
        elif base.startswith( strip ):
            return ctype_to_pytype( base[len(strip):] )
    if base.endswith( '*' ):
        new = ctype_to_pytype( base[:-1] )
        if new in ('_cs.GLvoid','_cs.void','_cs.c_void'):
            return 'ctypes.c_void_p'
        elif new == 'ctypes.c_void_p':
            return 'arrays.GLvoidpArray'
        elif new in CTYPE_TO_ARRAY_TYPE:
            return 'arrays.%s'%(CTYPE_TO_ARRAY_TYPE[new])
        elif new in ( 'arrays.GLcharArray','arrays.GLcharARBArray','arrays.GLbyteArray'):
            # can't have a pointer to these...
            return 'ctypes.POINTER( ctypes.POINTER( _cs.GLchar ))'
        elif new in ( '_cs.GLcharARB',):
            return 'ctypes.POINTER( ctypes.c_char_p )'
        else:
            log.debug( 'Unconverted pointer type: %r', new )
            return 'ctypes.POINTER(%s)'%(new)
    elif base in SPECIAL_TYPES:
        return SPECIAL_TYPES.get( base )
    else:
        if base == 'int':
            base = 'c_int'
        elif base == 'float':
            return 'ctypes.c_float'
        elif ' ' in base:
            components = base.split()
            if components[0] == 'unsigned':
                if len(components) == 2:
                    base = 'c_u'+components[1]
                else:
                    raise ValueError( base )
            else:
                raise ValueError( base )
        return '_cs.%s'%(base,)



#! /usr/bin/env python
"""Download and replicate the Win32+Win64 glut into our DLLs directory...

Visual Studio keeps eating my 64-bit configurations, so we'll use the binary distribution for now...
"""
import sys, os, subprocess, requests, logging, platform, shutil, glob
log = logging.getLogger( 'buildgle' )

DOWNLOAD_URL = 'http://www.idfun.de/glut64/glut-3.7.6-bin-32and64.zip'
GLUT_VERSION = '3.7.6'
GLUT_SOURCE_DIR = 'glut-%(GLUT_VERSION)s-bin'%globals()
ZIP_FILE = '%(GLUT_SOURCE_DIR)s-32and64.zip'%globals()

def download_and_unpack():
    if not os.path.exists( ZIP_FILE ):
        response = requests.get( DOWNLOAD_URL )
        if response.ok:
            open( ZIP_FILE, 'wb').write( response.content )
    if not os.path.exists( GLUT_SOURCE_DIR ):
        subprocess.check_call( 'unzip %(ZIP_FILE)s'%globals() )

def copy():
    current = os.getcwd()
    target_dir = os.path.join( current, '..','OpenGL','DLLS' )
    for fn in glob.glob( os.path.join( GLUT_SOURCE_DIR, '*.dll' )):
        shutil.copyfile( fn, os.path.join( target_dir, os.path.basename( fn )))
    shutil.copyfile( os.path.join( GLUT_SOURCE_DIR, 'README-win32.txt' ), os.path.join( target_dir, 'glut_README.txt' ))

def main():
    download_and_unpack()
    copy()

if __name__ == "__main__":
    logging.basicConfig( level=logging.INFO )
    main()



#!/usr/bin/python
from __future__ import print_function
import sys, os
import pygame
from pygame.locals import *
from OpenGL.GL import *
from OpenGL.GLU import *
import psutil


def main():
    pygame.init()
    pygame.display.set_mode((800, 600), pygame.OPENGL | pygame.DOUBLEBUF)
    glClearColor(1.0, 0.0, 0.0, 1.0)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

    done = False

    while not done:
        mem = 0
        for i in range(0, 500):
            if i == 10:
                mem = psutil.Process(os.getpid()).memory_percent()
            if i > 400:
                new_mem = psutil.Process(os.getpid()).memory_percent()
                assert new_mem == mem, (new_mem, mem)
                break

            modelview_matrix = glGetFloatv(GL_MODELVIEW_MATRIX)
            assert modelview_matrix is not None
        break
    sys.stdout.write('OK\n')
    sys.stdout.flush()


if __name__ == '__main__':
    main()



"""Pygame independent GL test decorator

Usage:

    from pygamegltest import gltest, pygamewrapper
    @gltest
    def function():
        '''Function that should run under a GL context
        
        Size will be (300,300) and title will be the name of the 
        function passed in.
        '''
        return None
    
    @pygamewrapper( size=(640,480), name='Cool Test' )
    def function():
        '''Function that should run under specially configure context'''
        return None

Will create a new Pygame OpenGL context and let you run the function 
in that context. Each invocation gets a new context, so you should 
not use this for large test suites.
"""
import pygame, pygame.display
from functools import wraps

SCREEN = None


def pygametest(size=(300, 300), name=None):
    def gltest(function):
        """Decorator to allow a function to run in a Pygame OpenGL context"""

        @wraps(function)
        def test_function(*args, **named):
            global SCREEN
            pygame.display.init()
            SCREEN = pygame.display.set_mode(
                size,
                pygame.OPENGL | pygame.DOUBLEBUF,
            )
            pygame.display.set_caption(name or function.__name__)
            pygame.key.set_repeat(500, 30)
            try:
                return function(*args, **named)
            finally:
                pygame.display.quit()
                pygame.quit()

        return test_function

    return gltest


gltest = pygametest()



#! /usr/bin/env python
from __future__ import print_function
import pygame, pygame.display
import logging, time, traceback, unittest, os, pytest

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)
HERE = os.path.dirname(__file__)
import pickle

try:
    import cPickle
except ImportError as err:
    cPickle = pickle

try:
    from numpy import *
except ImportError as err:
    array = None

pygame.display.init()
import OpenGL

if os.environ.get('TEST_NO_ACCELERATE'):
    OpenGL.USE_ACCELERATE = False
# OpenGL.FULL_LOGGING = True
OpenGL.CONTEXT_CHECKING = True
OpenGL.FORWARD_COMPATIBLE_ONLY = False
OpenGL.UNSIGNED_BYTE_IMAGES_AS_STRING = True

# from OpenGL._bytes import bytes, _NULL_8_BYTE, unicode, as_8_bit
from OpenGL.GL import *

try:
    glGetError()
except error.NoContext as err:
    # good, should have got this error
    pass
else:
    print('WARNING: Failed to catch invalid context')
    # raise RuntimeError( """Did not catch invalid context!""" )
from OpenGL import error
from OpenGL.GLU import *
import OpenGL
from OpenGL.extensions import alternate, GLQuerier
from OpenGL.GL.framebufferobjects import *
from OpenGL.GL.EXT.multi_draw_arrays import *
from OpenGL.GL.ARB.imaging import *
from OpenGL._bytes import _NULL_8_BYTE

glMultiDrawElements = alternate(
    glMultiDrawElementsEXT,
    glMultiDrawElements,
)
import basetestcase


class TestCore(basetestcase.BaseTest):
    def test_errors(self):
        """Test for error catching/checking"""
        try:
            glClear(GL_INVALID_VALUE)
        except Exception as err:
            assert err.err == 1281, ("""Expected invalid value (1281)""", err.err)
        else:
            assert not OpenGL.ERROR_CHECKING, """No error on invalid glClear"""
        try:
            glColorPointer(GL_INVALID_VALUE, GL_BYTE, 0, None)
        except Exception as err:
            assert err.err == 1281, ("""Expected invalid value (1281)""", err.err)
            assert err.baseOperation, err.baseOperation
            assert err.pyArgs == (GL_INVALID_VALUE, GL_BYTE, 0, None), err.pyArgs
            assert err.cArgs == (GL_INVALID_VALUE, GL_BYTE, 0, None), err.cArgs
        else:
            assert not OpenGL.ERROR_CHECKING, """No error on invalid glColorPointer"""
        try:
            glBitmap(-1, -1, 0, 0, 0, 0, as_8_bit(""))
        except Exception as err:
            assert err.err in (1281, 1282), (
                """Expected invalid value (1281) or invalid operation (1282)""",
                err.err,
            )
        else:
            assert not OpenGL.ERROR_CHECKING, """No error on invalid glBitmap"""

    if not OpenGL.ERROR_ON_COPY:

        def test_simple(self):
            """Test for simple vertex-based drawing"""
            glDisable(GL_LIGHTING)
            glBegin(GL_TRIANGLES)
            try:
                try:
                    glVertex3f(0.0, 1.0, 0.0)
                except Exception:
                    traceback.print_exc()
                glVertex3fv([-1, 0, 0])
                glVertex3dv([1, 0, 0])
                try:
                    glVertex3dv([1, 0, 4, 5])
                except ValueError:
                    # Got expected value error (good)
                    assert OpenGL.ARRAY_SIZE_CHECKING, """Should have raised ValueError when doing array size checking"""
                else:
                    assert not OpenGL.ARRAY_SIZE_CHECKING, """Should not have raised ValueError when not doing array size checking"""
            finally:
                glEnd()
            a = glGenTextures(1)
            assert a
            b = glGenTextures(2)
            assert len(b) == 2

    def test_arbwindowpos(self):
        """Test the ARB window_pos extension will load if available"""
        from OpenGL.GL.ARB.window_pos import glWindowPos2dARB

        if glWindowPos2dARB:
            glWindowPos2dARB(0.0, 3.0)

    def test_getstring(self):
        assert glGetString(GL_EXTENSIONS)

    def test_constantPickle(self):
        """Test that our constants can be pickled/unpickled properly"""
        for p in pickle, cPickle:
            v = p.loads(p.dumps(GL_VERTEX_ARRAY))
            assert v == GL_VERTEX_ARRAY, (v, GL_VERTEX_ARRAY)
            assert v.name == GL_VERTEX_ARRAY.name, v.name

    def test_nonFloatColor(self):
        """Test that we handle non-floating-point colour inputs"""
        for notFloat, shouldWork in ((0, True), (object(), False), (object, False)):
            try:
                glColor4f(0, 1, 1, notFloat)
            except Exception:
                if shouldWork:
                    raise
            else:
                if not shouldWork:
                    raise RuntimeError(
                        """Expected failure for non-float value %s, succeeded"""
                        % (notFloat,)
                    )

    if array:

        def test_arrayTranspose(self):
            m = glGetFloatv(GL_MODELVIEW_MATRIX)
            glMatrixMode(GL_MODELVIEW)
            glLoadIdentity()

            t = eye(4)
            t[3, 0] = 20.0

            # the following glMultMatrixf call ignored this transpose
            t = t.T
            if OpenGL.ERROR_ON_COPY:
                t = ascontiguousarray(t)

            glMultMatrixd(t)

            m = glGetFloatv(GL_MODELVIEW_MATRIX)
            assert allclose(m[-1], [0, 0, 0, 1]), m

    if array:
        # currently crashes in py_buffer operation, so reverted to raw numpy
        # api
        def test_mmap_data(self):
            """Test that we can use mmap data array

            If we had a reasonable lib that dumped raw image data to a shared-mem file
            we might be able to use this for movie display :)
            """
            fh = open('mmap-test-data.dat', 'wb+')
            fh.write(_NULL_8_BYTE * (32 * 32 * 3 + 1))
            fh.flush()
            fh.close()
            # using memmap here...
            data = memmap('mmap-test-data.dat')
            for i in range(0, 255, 2):
                glDrawPixels(
                    32,
                    32,
                    GL_RGB,
                    GL_UNSIGNED_BYTE,
                    data,
                )
                glFlush()
                pygame.display.flip()
                data[::2] = i
                time.sleep(0.001)

    if array:

        def test_vbo(self):
            """Test utility vbo wrapper"""
            from OpenGL.arrays import vbo

            if not vbo.get_implementation():
                return
            if not glVertexPointerd or not glDrawElements:
                return
            points = array(
                [
                    [0, 0, 0],
                    [0, 1, 0],
                    [1, 0.5, 0],
                    [1, 0, 0],
                    [1.5, 0.5, 0],
                    [1.5, 0, 0],
                ],
                dtype='d',
            )
            indices = array(
                range(len(points)),
                ['i', 'I'][int(bool(OpenGL.ERROR_ON_COPY))],  # test coercion if we can
            )
            d = vbo.VBO(points)
            glDisable(GL_CULL_FACE)
            glNormal3f(0, 0, 1)
            glColor3f(1, 1, 1)
            glEnableClientState(GL_VERTEX_ARRAY)
            try:
                for x in range(1, 255, 10):
                    d.bind()
                    try:
                        glVertexPointerd(d)
                        glDrawElements(
                            GL_LINE_LOOP, len(indices), GL_UNSIGNED_INT, indices
                        )
                    finally:
                        d.unbind()
                    lastPoint = array([[1.5, (1 / 255.0) * float(x), 0]])
                    d[-2:-1] = lastPoint
                    glFlush()
                    pygame.display.flip()
                    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
                    time.sleep(0.001)
            finally:
                glDisableClientState(GL_VERTEX_ARRAY)
            # bug report from Dan Helfman, delete shouldn't cause
            # errors if called explicitly
            d.delete()

        def test_glgetbufferparameter(self):
            if not glGenBuffers or not glGenVertexArrays:
                return None
            buffer = glGenBuffers(1)
            vertex_array = glGenVertexArrays(1, buffer)
            glBindBuffer(GL_ARRAY_BUFFER, buffer)
            try:
                mapped = glGetBufferParameteriv(GL_ARRAY_BUFFER, GL_BUFFER_MAPPED)
                assert mapped == (
                    GL_FALSE if OpenGL.SIZE_1_ARRAY_UNPACK else [GL_FALSE]
                ), mapped
            finally:
                glBindBuffer(GL_ARRAY_BUFFER, 0)
                glDeleteVertexArrays(1, vertex_array)
                glDeleteBuffers(1, buffer)

        @pytest.mark.skipif(
            not OpenGL.SIZE_1_ARRAY_UNPACK,
            reason="Tests array unpack that is not configured",
        )
        def test_scalars_mapped_to_arrays(self):
            if not glGenBuffers or not glGenVertexArrays:
                return None
            buffer = glGenBuffers(1)
            assert isscalar(buffer), type(buffer)
            glDeleteBuffers(1, buffer)

    def test_glbufferparameter_create(self):
        if not glGenBuffers or not glGenVertexArrays:
            return None
        for create in [True, False]:
            buffer = glGenBuffers(1)
            vertex_array = glGenVertexArrays(1, buffer)
            glBindBuffer(GL_ARRAY_BUFFER, buffer)
            try:
                for param, expected in [
                    (GL_BUFFER_SIZE, 0),
                    (GL_BUFFER_MAPPED, GL_FALSE),
                    (GL_BUFFER_STORAGE_FLAGS, 0),
                    (GL_BUFFER_USAGE, GL_STATIC_DRAW),
                ]:
                    if create:
                        mapped = GLint(-1)
                        glGetBufferParameteriv(GL_ARRAY_BUFFER, param, mapped)
                        assert mapped.value == expected, (param, mapped, expected)
                    else:
                        mapped = glGetBufferParameteriv(GL_ARRAY_BUFFER, param)
                        if param == GL_BUFFER_USAGE:
                            assert mapped[0] == expected, (param, mapped, expected)
                        else:
                            if OpenGL.SIZE_1_ARRAY_UNPACK:
                                assert mapped == expected, (param, mapped, expected)
                            else:
                                assert mapped[0] == expected, (
                                    param,
                                    mapped[0],
                                    expected,
                                )
            finally:
                glBindBuffer(GL_ARRAY_BUFFER, 0)
                glDeleteVertexArrays(1, vertex_array)
                glDeleteBuffers(1, buffer)

    def test_fbo(self):
        """Test that we support framebuffer objects

        http://www.gamedev.net/reference/articles/article2331.asp
        """
        if not glGenFramebuffers:
            print('No Frame Buffer Support!')
            return False
        width = height = 128
        fbo = glGenFramebuffers(1)
        glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        depthbuffer = glGenRenderbuffers(1)
        glBindRenderbuffer(GL_RENDERBUFFER, depthbuffer)
        glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, width, height)
        glFramebufferRenderbuffer(
            GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, depthbuffer
        )

        img = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, img)
        # NOTE: these lines are *key*, without them you'll likely get an unsupported format error,
        # ie. GL_FRAMEBUFFER_UNSUPPORTED
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexImage2D(
            GL_TEXTURE_2D,
            0,
            GL_RGB8,
            width,
            height,
            0,
            GL_RGB,
            GL_INT,
            None,  # no data transferred
        )
        glFramebufferTexture2D(
            GL_FRAMEBUFFER,
            GL_COLOR_ATTACHMENT0,
            GL_TEXTURE_2D,
            img,
            0,  # mipmap level, normally 0
        )
        status = glCheckFramebufferStatus(GL_FRAMEBUFFER)
        assert status == GL_FRAMEBUFFER_COMPLETE, status
        glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        glPushAttrib(GL_VIEWPORT_BIT)  # viewport is shared with the main context
        try:
            glViewport(0, 0, width, height)

            # rendering to the texture here...
            glColor3f(1, 0, 0)
            glNormal3f(0, 0, 1)
            glBegin(GL_QUADS)
            for v in [[0, 0, 0], [0, 1, 0], [1, 1, 0], [1, 0, 0]]:
                glColor3f(*v)
                glVertex3d(*v)
            glEnd()
        finally:
            glPopAttrib()
            # restore viewport
        glBindFramebuffer(GL_FRAMEBUFFER, 0)  # unbind

        glBindTexture(GL_TEXTURE_2D, img)

        glEnable(GL_TEXTURE_2D)

        # rendering with the texture here...
        glColor3f(1, 1, 1)
        glNormal3f(0, 0, 1)
        glDisable(GL_LIGHTING)
        glBegin(GL_QUADS)
        try:
            for v in [[0, 0, 0], [0, 1, 0], [1, 1, 0], [1, 0, 0]]:
                glTexCoord2f(*v[:2])
                glVertex3d(*v)
        finally:
            glEnd()

    def test_gl_1_2_support(self):
        if glBlendColor:
            glBlendColor(0.3, 0.4, 1.0, 0.3)
            print('OpenGL 1.2 support')

    if array:

        def test_glmultidraw(self):
            """Test that glMultiDrawElements works, uses glDrawElements"""
            if glMultiDrawElements:
                points = array(
                    [(i, 0, 0) for i in range(8)] + [(i, 1, 0) for i in range(8)], 'd'
                )
                indices = array(
                    [
                        [
                            0,
                            8,
                            9,
                            1,
                            2,
                            10,
                            11,
                            3,
                        ],
                        [4, 12, 13, 5, 6, 14, 15, 7],
                    ],
                    'B',
                )
                index_pointers = arrays.GLvoidpArray.zeros((2,))
                index_pointers[0] = arrays.GLbyteArray.dataPointer(indices)
                index_pointers[1] = arrays.GLbyteArray.dataPointer(indices[1])
                counts = [len(x) for x in indices]
                if OpenGL.ERROR_ON_COPY:
                    counts = (GLuint * len(counts))(*counts)
                glEnableClientState(GL_VERTEX_ARRAY)
                glDisableClientState(GL_COLOR_ARRAY)
                glDisableClientState(GL_NORMAL_ARRAY)
                try:
                    glVertexPointerd(points)
                    glDisable(GL_LIGHTING)
                    try:
                        glMultiDrawElements(
                            GL_QUAD_STRIP, counts, GL_UNSIGNED_BYTE, index_pointers, 2
                        )
                    finally:
                        glEnable(GL_LIGHTING)
                finally:
                    glDisableClientState(GL_VERTEX_ARRAY)
            else:
                print('No multi_draw_arrays support')

    def test_glDrawBuffers_list(self):
        """Test that glDrawBuffers with list argument doesn't crash"""
        if not glDrawBuffers:
            return
        a_type = GLenum * 2
        args = a_type(
            GL_COLOR_ATTACHMENT0,
            GL_COLOR_ATTACHMENT1,
        )
        try:
            glDrawBuffers(2, args)
        except GLError as err:
            assert err.err == GL_INVALID_OPERATION, err

    def test_glDrawBuffers_list_valid(self):
        """Test that glDrawBuffers with list argument where value is set"""
        if not glGenFramebuffers:
            return
        if not glDrawBuffers:
            return
        previous = glGetIntegerv(GL_READ_BUFFER)
        fbo = glGenFramebuffers(1)
        glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        try:
            img1, img2 = glGenTextures(2)
            for img in img1, img2:
                glBindTexture(GL_TEXTURE_2D, img)
                glTexImage2D(
                    GL_TEXTURE_2D,
                    0,
                    GL_RGB8,
                    300,
                    300,
                    0,
                    GL_RGB,
                    GL_INT,
                    None,  # no data transferred
                )

            glFramebufferTexture2D(
                GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, img1, 0
            )
            glFramebufferTexture2D(
                GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, img2, 0
            )
            a_type = GLenum * 2
            drawingBuffers = a_type(
                GL_COLOR_ATTACHMENT0,
                GL_COLOR_ATTACHMENT1,
            )
            glDrawBuffers(2, drawingBuffers)
            try:
                checkFramebufferStatus()
            except error.GLError:
                pass
            else:
                glReadBuffer(GL_COLOR_ATTACHMENT1)
                pixels = glReadPixels(0, 0, 10, 10, GL_RGB, GL_UNSIGNED_BYTE)
                assert len(pixels) == 300, len(pixels)
        finally:
            glBindFramebuffer(GL_FRAMEBUFFER, 0)

        glReadBuffer(previous)

    def test_get_version(self):
        from OpenGL.extensions import hasGLExtension

        if hasGLExtension('GL_VERSION_GL_2_0'):
            assert glShaderSource
            assert glUniform1f
        else:
            assert not glShaderSource
            assert not glUniform1f

    def test_lookupint(self):
        from OpenGL.raw.GL import _lookupint

        if GLQuerier.pullVersion() < [2]:
            return
        l = _lookupint.LookupInt(GL_NUM_COMPRESSED_TEXTURE_FORMATS, GLint)
        result = int(l)
        if not os.environ.get('TRAVIS'):
            assert (
                result
            ), "No compressed textures on this platform? that seems unlikely"
        else:
            "Travis xvfb doesn't normally have compressed textures, possible upgrade?"

    def test_glget(self):
        """Test that we can run glGet... on registered constants without crashing..."""
        from OpenGL.raw.GL import _glgets

        get_items = sorted(_glgets._glget_size_mapping.items())
        for key, value in get_items:
            # There are glGet values that will cause a crash during cleanup
            # GL_ATOMIC_COUNTER_BUFFER_BINDING 0x92c1 crashes/segfaults
            if key >= 0x8DF9 and key <= 0x8E23:
                continue
            if key >= 0x92BE and key <= 0x92C9:
                continue
            # print( 'Trying glGetFloatv( 0x%x )'%(key,))
            try:
                result = glGetFloatv(key)
            except error.GLError as err:
                if err.err == GL_INVALID_ENUM:
                    pass
                elif err.err == GL_INVALID_OPERATION:
                    if key == 0x882D:  # gl draw buffer
                        pass
                else:
                    raise
            else:
                if value == (1,) and OpenGL.SIZE_1_ARRAY_UNPACK:
                    try:
                        result = float(result)
                    except TypeError as err:
                        err.args += (result, key)
                        raise
                else:
                    assert ArrayDatatype.dimensions(result) == value, (
                        result,
                        ArrayDatatype.dimensions(result),
                        value,
                    )

    def test_max_compute_work_group_invocations(self):
        from OpenGL.extensions import hasGLExtension

        if hasGLExtension('GL_ARB_compute_shader'):
            assert glGetIntegerv(GL_MAX_COMPUTE_WORK_GROUP_INVOCATIONS)

    def test_glCallLists_twice2(self):
        """SF#2829309 report that glCallLists doubles operation"""
        glRenderMode(GL_RENDER)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(40.0, 1.0, 1.0, 10.0)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        glTranslatef(0, 0, -3)
        first = glGenLists(2)
        second = first + 1

        glNewList(first, GL_COMPILE_AND_EXECUTE)
        glInitNames()
        if not OpenGL.ERROR_ON_COPY:
            glCallLists([second])  # replace with gCallList(2)
        else:
            lists = (GLuint * 1)()
            lists[0] = second
            glCallLists(lists)
        # glCallList(second)
        glEndList()

        glNewList(second, GL_COMPILE)
        glPushName(1)
        glBegin(GL_POINTS)
        glVertex3f(0, 0, 0)
        glEnd()
        glEndList()
        glCallList(second)
        glPopName()
        depth = glGetIntegerv(GL_NAME_STACK_DEPTH)
        assert depth in (
            0,
            (0,),
        ), depth  # have popped, but even then, were' not in the mode...

        glSelectBuffer(100)
        glRenderMode(GL_SELECT)
        glCallList(1)
        depth = glGetIntegerv(GL_NAME_STACK_DEPTH)
        assert depth in (1, (1,)), depth  # should have a single record
        glPopName()
        records = glRenderMode(GL_RENDER)
        # reporter says sees two records, Linux sees none, Win32 sees 1 :(
        assert len(records) == 1, records

    def test_orinput_handling(self):
        if not glGenVertexArrays:
            return None
        x = glGenVertexArrays(1)
        x = int(x)  # check that we got x as an integer-compatible value
        x2 = GLuint()
        r_value = glGenVertexArrays(1, x2)
        assert x2.value, x2.value
        assert r_value

        color = glGetFloatv(GL_FOG_COLOR)
        color2 = (GLfloat * 4)()
        glGetFloatv(GL_FOG_COLOR, color2)
        for a, b in zip(color, color2):
            assert a == b, (color, color2)

    def test_get_read_fb_binding(self):
        if GLQuerier.pullVersion() < [3]:
            return
        glGetInteger(GL_READ_FRAMEBUFFER_BINDING)

    def test_shader_compile_string(self):
        if not glCreateShader:
            return None
        shader = glCreateShader(GL_VERTEX_SHADER)

        def glsl_version():
            """Parse GL_SHADING_LANGUAGE_VERSION into [int(major),int(minor)]"""
            version = glGetString(GL_SHADING_LANGUAGE_VERSION)
            version = version.split(as_8_bit(' '))[0]
            version = [int(x) for x in version.split(as_8_bit('.'))[:2]]
            return version

        if glsl_version() < [3, 3]:
            return
        SAMPLE_SHADER = '''#version 330
        void main() { gl_Position = vec4(0,0,0,0);}'''
        if OpenGL.ERROR_ON_COPY:
            SAMPLE_SHADER = as_8_bit(SAMPLE_SHADER)
        glShaderSource(shader, SAMPLE_SHADER)
        glCompileShader(shader)
        if not bool(glGetShaderiv(shader, GL_COMPILE_STATUS)) == True:
            print('Info log:')
            print(glGetShaderInfoLog(shader))
            assert False, """Failed to compile"""

    def test_gen_framebuffers_twice(self):
        if not glGenFramebuffersEXT:
            return
        glGenFramebuffersEXT(1)
        f1 = glGenFramebuffersEXT(1)
        f2 = glGenFramebuffersEXT(1)
        assert f1 != f2, (f1, f2)
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, f2)
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)

    def test_compressed_data(self):
        from OpenGL.extensions import hasGLExtension

        if hasGLExtension('GL_EXT_texture_compression_s3tc'):
            from OpenGL.GL.EXT import texture_compression_s3tc as s3tc

            texture = glGenTextures(1)
            glEnable(GL_TEXTURE_2D)
            image_type = GLubyte * 256 * 256
            image = image_type()
            glCompressedTexImage2D(
                GL_TEXTURE_2D,
                0,
                s3tc.GL_COMPRESSED_RGBA_S3TC_DXT5_EXT,
                256,
                256,
                0,
                # 256*256*1,
                image,
            )
            assert texture


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    unittest.main()
    pygame.display.quit()
    pygame.quit()



"""Test simple functions (i.e. no pointers involved)"""

from __future__ import print_function
import sys
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from OpenGL._bytes import as_8_bit
import time

start = time.time()

window = None


def drawText(value, x, y, windowHeight, windowWidth, step=18):
    """Draw the given text at given 2D position in window"""
    glMatrixMode(GL_PROJECTION)
    # For some reason the GL_PROJECTION_MATRIX is overflowing with a single push!
    # glPushMatrix()
    matrix = glGetDouble(GL_PROJECTION_MATRIX)

    glLoadIdentity()
    glOrtho(0.0, windowHeight or 32, 0.0, windowWidth or 32, -1.0, 1.0)
    glMatrixMode(GL_MODELVIEW)
    glPushMatrix()
    glLoadIdentity()
    glRasterPos2i(x, y)
    lines = 0
    for character in value:
        if character == '\n':
            glRasterPos2i(x, y - (lines * 18))
        else:
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, ord(character))
    glPopMatrix()
    glMatrixMode(GL_PROJECTION)
    # For some reason the GL_PROJECTION_MATRIX is overflowing with a single push!
    # glPopMatrix();
    glLoadMatrixd(matrix)  # should have un-decorated alias for this...

    glMatrixMode(GL_MODELVIEW)


def display():
    glutSetWindow(window)
    glClearColor(0.0, 0.0, (time.time() % 1.0) / 1.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT)
    drawText('hello', 20, 20, size[0], size[1])
    # glutBitmapCharacter( GLUT_BITMAP_8_BY_13, ord('a'))
    glutSolidTeapot(0.2)
    glFlush()
    glutSwapBuffers()


size = (250, 250)


def reshape(*args):
    global size
    size = args
    glViewport(*((0, 0) + args))
    display()


def ontimer(*args):
    print(('timer', args, '@time', time.time() - start))
    glutTimerFunc(1000, ontimer, 24)


def idle():
    delta = time.time() - start
    if delta < 2:
        global size
        x, y = size
        if delta < 5:
            change = +1
        else:
            change = -1
        x = x - change
        y = y + change
        if x < 1:
            x = 1
        if y < 1:
            y = 1
        glutReshapeWindow(x, y)
        size = (x, y)
        glutSetWindow(window)
        glutPostRedisplay()
    else:
        sys.stdout.write("OK\n")
        sys.stdout.flush()
        glutDestroyWindow(window)
        try:
            if fgDeinitialize:
                fgDeinitialize(False)
        except NameError:
            pass  # Older PyOpenGL, you may see a seg-fault here...

        os._exit(0)


def printFunction(name):
    def onevent(*args):
        print(('%s -> %s' % (name, ", ".join([str(a) for a in args]))))

    return onevent


if __name__ == "__main__":
    import sys

    newArgv = glutInit(sys.argv)
    print(('newArguments', newArgv))
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB)
    glutInitWindowSize(250, 250)
    glutInitWindowPosition(100, 100)
    window = glutCreateWindow(as_8_bit("hello"))
    print(('window', repr(window)))
    glutDisplayFunc(display)
    glutReshapeFunc(reshape)
    glutMouseFunc(printFunction('Mouse'))
    glutEntryFunc(printFunction('Entry'))
    glutKeyboardFunc(printFunction('Keyboard'))
    glutKeyboardUpFunc(printFunction('KeyboardUp'))
    glutMotionFunc(printFunction('Motion'))
    glutPassiveMotionFunc(printFunction('PassiveMotion'))
    glutVisibilityFunc(printFunction('Visibility'))
    glutWindowStatusFunc(printFunction('WindowStatus'))
    glutSpecialFunc(printFunction('Special'))
    glutSpecialUpFunc(printFunction('SpecialUp'))
    glutTimerFunc(1000, ontimer, 23)

    glutIdleFunc(idle)
    glutMainLoop()



"""Test GLUT forward-compatible mode..."""

from __future__ import print_function
import OpenGL

OpenGL.FORWARD_COMPATIBLE_ONLY = True
OpenGL.ERROR_CHECKING = True
# OpenGL.USE_ACCELERATE = False
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *

# from OpenGL.GL import shaders
import time, sys, os

start = time.time()

window = None


def display():
    try:
        glutSetWindow(window)
        glClearColor(0.0, 0.0, (time.time() % 1.0) / 1.0, 0.0)
        glClear(GL_COLOR_BUFFER_BIT)
        try:
            glGetString(GL_EXTENSIONS)
        except GLError:
            pass
        else:
            print('Egads, glGetString should not have worked!')
        assert bool(glGenVertexArrays), "Should have vertex array support in 3.2"
        glFlush()
        glutSwapBuffers()
    except Exception:
        glutDestroyWindow(window)
        raise


size = (250, 250)


def reshape(*args):
    global size
    size = args
    glViewport(*((0, 0) + args))
    display()


def ontimer(*args):
    #    print 'timer', args, '@time', time.time()-start
    glutTimerFunc(1000, ontimer, 24)


def idle():
    delta = time.time() - start
    if delta < 2:
        global size
        x, y = size
        if delta < 5:
            change = +1
        else:
            change = -1
        x = x - change
        y = y + change
        if x < 1:
            x = 1
        if y < 1:
            y = 1
        glutReshapeWindow(x, y)
        size = (x, y)
        glutSetWindow(window)
        glutPostRedisplay()
    else:
        glutDestroyWindow(window)
        sys.stdout.write('OK\n')
        sys.stdout.flush()
        try:
            if fgDeinitialize:
                fgDeinitialize(False)
        except NameError:
            pass  # Older PyOpenGL, you may see a seg-fault here...

        os._exit(0)


def printFunction(name):
    def onevent(*args):
        print('%s -> %s' % (name, ", ".join([str(a) for a in args])))

    return onevent


if __name__ == "__main__":
    import sys

    newArgv = glutInit(sys.argv)
    glutInitContextVersion(3, 1)
    glutInitContextFlags(GLUT_FORWARD_COMPATIBLE)
    glutInitContextProfile(GLUT_CORE_PROFILE)
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH)

    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS)
    glutInitWindowSize(250, 250)
    glutInitWindowPosition(100, 100)
    window = glutCreateWindow("hello")
    print('window', repr(window))
    glutDisplayFunc(display)
    glutReshapeFunc(reshape)
    glutMouseFunc(printFunction('Mouse'))
    glutEntryFunc(printFunction('Entry'))
    glutKeyboardFunc(printFunction('Keyboard'))
    glutKeyboardUpFunc(printFunction('KeyboardUp'))
    glutMotionFunc(printFunction('Motion'))
    glutPassiveMotionFunc(printFunction('PassiveMotion'))
    glutVisibilityFunc(printFunction('Visibility'))
    glutWindowStatusFunc(printFunction('WindowStatus'))
    glutSpecialFunc(printFunction('Special'))
    glutSpecialUpFunc(printFunction('SpecialUp'))
    glutTimerFunc(1000, ontimer, 23)

    glutIdleFunc(idle)
    glutMainLoop()



import logging

logging.basicConfig(level=logging.INFO)
logging.getLogger('OpenGL.plugins').setLevel(logging.ERROR)


def fail_handler(*args, **named):
    raise RuntimeError("Should not have called this handler!")


logging.getLogger('OpenGL.plugins').addHandler(fail_handler)
from OpenGL.arrays import arraydatatype

dt = arraydatatype.GLuintArray
d = dt.zeros((3,))

print('OK')



#! /usr/bin/env python
"""Registry for loading Khronos API definitions from XML files"""
from lxml import etree as ET
import os, sys, json, logging
from OpenGL._bytes import as_8_bit, unicode, as_str

log = logging.getLogger(__name__)
HERE = os.path.dirname(__file__)

LENGTH_OVERRIDES = {
    'glGetPolygonStipple': {
        'mask': str(32 * 32 / 8),  # 32x32 bits
    },
    'glGetUniformfv': {
        'params': None,
    },
    'glGetUniformiv': {
        'params': None,
    },
    #    'glShaderSourceARB': {
    #        'string': None,
    #    },
    #    'glShaderSource': {
    #        'string': None,
    #    },
}


class Registry(object):
    def __init__(self):
        self.type_set = {}
        self.enum_namespaces = {}
        self.enum_groups = {}
        self.enumeration_set = {}
        self.command_set = {}
        self.apis = {}
        self.feature_set = {}
        self.extension_set = {}
        self.output_mapping = json.loads(
            open(os.path.join(HERE, 'gl_out_parameters.json')).read()
        )
        self.output_enum_groups = {}

    def load(self, tree):
        """Load an lxml.etree structure into our internal descriptions"""
        self.dispatch(tree, None)

    def dispatch(self, tree, context=None):
        """Dispatch for all children of the element"""
        for element in tree:
            if isinstance(element.tag, (str, unicode)):
                method = getattr(self, element.tag, None)
                if method:
                    method(element, context)
                else:
                    print('Expand', element.tag)
                    self.dispatch(element, context)

    def type(self, element, context=None):
        name = element.get('name')
        if not name:
            name = element.find('name').text
        self.type_set[as_str(name)] = element

    def debug_types(self):
        for name, type in self.types.items():
            print(name, type)

    def enums(self, element, context=None):
        name = as_str(element.get('namespace'))
        if name not in self.enum_namespaces:
            namespace = EnumNamespace(name)
            self.enum_namespaces[name] = namespace
        else:
            namespace = self.enum_namespaces[name]
        self.dispatch(element, namespace)

    def enum(self, element, context=None):
        if isinstance(context, EnumNamespace):
            name, value = as_str(element.get('name')), element.get('value')
            enum = Enum(name, value)
            context.append(enum)
            self.enumeration_set[name] = enum
        elif isinstance(context, (Require, Remove)):
            context.append(self.enumeration_set[element.get('name')])
        elif isinstance(context, EnumGroup):
            name = element.get('name')
            assert name, 'No name on %s' % ET.tostring(element)
            context.append(as_str(name))

    def debug_enums(self):
        for name, namespace in self.enum_namespaces.items():
            print('Namespace', namespace.namespace)
            for enum in namespace:
                print('  ', enum)

    def command(self, element, context=None):
        """Parse command definition into structured format"""
        proto = element.find('proto')
        if proto is not None:
            name = as_str(proto.find('name').text)
            assert name, 'No name in command: %s' % (ET.tostring(element))
            return_type = self._type_decl(proto)
            assert return_type, 'No return type in command: %s' % (ET.tostring(element))
            arg_names = []
            arg_types = []
            lengths = {}
            groups = {}
            for param in [x for x in element if x.tag == 'param']:
                pname = as_str(param.find('name').text)
                arg_names.append(pname)
                arg_types.append(self._type_decl(param))
                if param.get('len'):
                    length = param.get('len')
                    while length.endswith('*1'):
                        length = length[:-2]
                    length = LENGTH_OVERRIDES.get(name, {}).get(pname, length)
                    lengths[pname] = length
                if param.get('group'):
                    groups[pname] = param.get('group')
            aliases = []
            for alias in [x for x in element if x.tag == 'alias']:
                aliases.append(alias.get('name'))
            # Process lengths to look for output parameters
            outputs = self.output_mapping.get(name)
            command = Command(
                name,
                return_type,
                arg_names,
                arg_types,
                aliases,
                lengths,
                groups,
                outputs=outputs,
            )
            self.command_set[name] = command
        elif isinstance(context, (Require, Remove)):
            context.append(self.command_set[element.get('name')])

    def _type_decl(self, proto):
        """Get the string type declaration for parent (proto/param)"""
        return_type = []
        if proto.text:
            return_type.append(proto.text)
        for item in proto:
            if item.tag == 'name':
                break
            else:
                if item.text:
                    return_type.append(item.text.strip())
                if item.tail:
                    return_type.append(item.tail.strip())
        return ' '.join([x for x in return_type if x]) or 'void'

    def debug_commands(self):
        for name, command in sorted(self.command_set.items()):
            print(command)

    def feature(self, element, context=None):
        api, name, number = [element.get(x) for x in ('api', 'name', 'number')]
        feature = Feature(api, name, number)
        self.feature_set[name] = feature
        self.dispatch(element, feature)

    def extension(self, element, context=None):
        name, apis, require = [element.get(x) for x in ['name', 'supported', 'protect']]
        extension = Extension(name, apis.split('|'), require)
        self.extension_set[name] = extension
        self.dispatch(element, extension)

    def unused(self, element, context=None):
        pass

    def group(self, element, context=None):
        name = element.get('name')
        current = self.enum_groups.get(name)
        if current is None:
            current = self.enum_groups[name] = EnumGroup(name)
        self.dispatch(element, current)

    def require(self, element, context):
        if isinstance(context, (Feature, Extension)):
            profile, comment = element.get('profile'), element.get('comment')
            require = Require(profile, comment)
            context.append(require)
            self.dispatch(element, require)

    def remove(self, element, context):
        if isinstance(context, Feature):
            profile, comment = element.get('profile'), element.get('comment')
            remove = Remove(profile, comment)
            context.append(remove)
            self.dispatch(element, remove)

    def debug_apis(self):
        print([x.api for x in self.feature_set.values()])


class EnumNamespace(list):
    def __init__(self, namespace, *args):
        self.namespace = namespace
        super(EnumNamespace, self).__init__(*args)


class EnumGroup(list):
    def __init__(self, name, *args):
        self.name = name
        super(EnumGroup, self).__init__(*args)


class Enum(object):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return '%s = %s' % (self.name, self.value)


class Command(object):
    def __init__(
        self,
        name,
        returnType,
        argNames,
        argTypes,
        aliases=None,
        lengths=None,
        groups=None,
        outputs=None,
    ):
        self.name = name
        self.returnType = returnType
        self.argNames = argNames
        self.argTypes = argTypes
        self.aliases = aliases or []
        self.lengths = lengths or {}
        self.groups = groups or {}
        self.outputs = outputs or {}
        self.output_groups = {}
        self.size_dependencies = self.calculate_sizes()

    def __repr__(self):
        return '%s %s( %s )' % (
            self.returnType,
            self.name,
            ', '.join(
                [
                    '%s %s' % (typ, name)
                    for (typ, name) in zip(self.argTypes, self.argNames)
                ]
            ),
        )

    def calculate_sizes(self):
        result = []
        other_lengths = self.lengths.copy()
        for target in self.outputs.keys():
            definition = self.lengths.get(target)
            if definition is None:
                if target not in self.argNames:
                    # may be a discrepency between .spec and xml registry file...
                    if target == 'params' and 'data' in self.argNames:
                        target = 'data'
                        definition = self.lengths.get('data')
            if target in other_lengths:
                del other_lengths[target]

            if definition is None:
                result.append((target, Output()))
            elif definition.startswith('COMPSIZE'):
                variables = definition[8:].strip('()').split(',')
                output_groups = {}
                if len(variables) == 1:
                    # for now we only support automated single-dependency wrapping...
                    for var in variables:
                        if var in self.groups:
                            output_groups.setdefault(self.groups[var], []).append(
                                target
                            )
                    self.output_groups.update(output_groups)
                result.append((target, Compsize(variables, output_groups)))
            elif definition.isdigit():
                result.append((target, Staticsize(int(definition, 10))))
            elif '*' in definition:
                var, multiple = definition.split('*')
                result.append((target, Multiple(var, int(multiple, 10))))
            else:
                result.append((target, Dynamicsize(definition)))
        for target, length in other_lengths.items():
            if length is None:
                # length/array-conversion suppressed
                continue
            if length.isdigit():
                result.append((target, StaticInput(int(length, 10))))
            elif length.startswith('COMPSIZE'):
                result.append((target, Input(length[9:-1])))
            elif length in self.argNames:
                result.append((target, DynamicInput(length)))
            elif '*' in length:
                params = [x.strip() for x in length.split('*')]
                in_set = [x for x in params if x in self.argNames]
                result.append((target, MultiplyInput(params)))
            elif '/' in length:
                params = [x.strip() for x in length.split('/')]
                in_set = [x for x in params if x in self.argNames]
                result.append((target, DivideInput(params)))
            else:
                raise RuntimeError((target, length))
        return dict(result)


class IsInput(object):
    pass


class Input(IsInput, object):
    """Unsized Input Parameter"""

    def __init__(self, value=None):
        self.value = value

    def __repr__(self):
        return repr(self.value)


class StaticInput(IsInput, int):
    """Statically sized input parameter"""


class DynamicInput(IsInput, str):
    """Dynamically sized based on other parameter"""


class MultiplyInput(IsInput, list):
    """Size depends on multiple elements being multiplied"""

    def __str__(self):
        return '*'.join(self)


class DivideInput(IsInput, list):
    """Size depends on multiple elements being multiplied"""

    def __str__(self):
        return '/'.join(self)


class Output(object):
    """Unsized output parameter"""


class Compsize(list):
    """Compute size based on other variables"""

    def __init__(self, iterable, groups=None):
        super(Compsize, self).__init__(iterable)
        self.groups = groups


class Staticsize(int):
    """Static output array size"""


class Dynamicsize(str):
    """Sized by the value in dynamic variable"""


class Multiple(list):
    """Variable * static size for array"""


# The order-dependent set of require/remove holding features/extensions
class Module(list):
    """Base class for Features and Extensions"""

    feature = False

    def __init__(self, name):
        self.name = name

    def members(self, of_type=None):
        members = []
        for req in self:
            if req.require:
                for item in req:
                    if of_type is not None:
                        if isinstance(item, of_type):
                            members.append(item)
                    else:
                        members.append(item)
        return members

    def enums(self):
        return self.members(Enum)

    def commands(self):
        return self.members(Command)


class Feature(Module):
    feature = True
    NORMALIZERS = {
        'GL_VERSION_ES_CM_1_0': 'GLES_VERSION_1_0',
        'GL_ES_VERSION_2_0': 'GLES_VERSION_2_0',
        'GL_ES_VERSION_3_0': 'GLES_VERSION_3_0',
    }

    def __init__(self, api, name, number):
        super(Feature, self).__init__(self.NORMALIZERS.get(name, name))
        self.api = api
        if name == 'GL_ES_VERSION_3_0':
            self.api = 'gles3'
        self.number = number

    _profiles = None

    @property
    def profiles(self):
        """Create set of profiles with subsets of our functionality"""
        if self._profiles is None:
            profiles = {}
            for req in self:
                # Logic isn't right here, there's a base and then
                # a set of profiles which customize the base...
                profile = req.profile or ''
                set = profiles.get(profile)
                if set is None:
                    set = Module(profile or '')
                    set.feature = True
                    profiles[profile] = set
                if req.require:

                    set.extend(req)
                else:
                    for item in req:
                        while item in set:
                            set.remove(item)
            self._profiles = sorted(profiles.values(), key=lambda x: x.name)
        return self._profiles


class Extension(Module):
    def __init__(self, name, apis, require=None):
        super(Extension, self).__init__(name)
        self.apis = apis  # only available for these APIs
        self.require = require

    @property
    def profiles(self):
        module = Module('default')
        module.extend(self)
        return module


class Require(list):
    require = True
    remove = False

    def __init__(self, profile=None, comment=None):
        self.profile = profile
        self.comment = comment
        super(Require, self).__init__()


class Remove(list):
    require = False
    remove = True

    def __init__(self, profile=None, comment=None):
        self.profile = profile
        self.comment = comment
        super(Remove, self).__init__()


def parse(xmlfile):
    registry = Registry()
    registry.load(ET.fromstring(open(xmlfile, 'rb').read()))
    return registry


if __name__ == "__main__":
    if sys.argv[1:]:
        for file in sys.argv[1:]:
            print(file)
            registry = parse(file)

    # registry.debug_types()
    # registry.debug_enums()
    # registry.debug_commands()
    # registry.debug_apis()



#! /usr/bin/env python
"""Attempt to import GLES libraries"""
import os

if not 'PYOPENGL_PLATFORM' in os.environ:
    os.environ['PYOPENGL_PLATFORM'] = 'egl'
from OpenGL.GLES1 import *
from OpenGL.GLES2 import *
from OpenGL.GLES3 import *

print('OK')



from __future__ import print_function

# import OpenGL
# OpenGL.USE_ACCELERATE=False
from OpenGL.GL import *
import sys, os

if (not sys.platform.startswith('linux')) or os.environ.get('XDG_SESSION_TYPE') == 'wayland':
    import pytest

    pytest.skip("Skipping GLX tests on non-GLX platforms", allow_module_level=True)
from OpenGL.GLX import *
from OpenGL.GLX.EXT.texture_from_pixmap import *

import os

attributes = [
    #    GLX_BIND_TO_TEXTURE_RGBA_EXT, 1,
    #    GLX_DRAWABLE_TYPE, GLX_PIXMAP_BIT,
    #    GLX_BIND_TO_TEXTURE_TARGETS_EXT, GLX_TEXTURE_2D_BIT_EXT,
    GLX_DOUBLEBUFFER,
    1,
    #    GLX_Y_INVERTED_EXT, GLX_DONT_CARE,
    GL_NONE,
]

# from OpenGL import platform
import ctypes
from OpenGL.platform import ctypesloader

X11 = ctypesloader.loadLibrary(ctypes.cdll, 'X11')
XDefaultScreen = X11.XDefaultScreen
XDefaultScreen.argtypes = [ctypes.POINTER(Display)]
XOpenDisplay = X11.XOpenDisplay
XOpenDisplay.restype = ctypes.POINTER(Display)
XDefaultScreenOfDisplay = X11.XDefaultScreenOfDisplay
XDefaultScreenOfDisplay.argtypes = [ctypes.POINTER(Display)]


def main():
    dsp = XOpenDisplay(os.environ.get('DISPLAY'))
    if not dsp:
        print("Could not get the display", os.environ.get('DISPLAY'))
        dsp = XOpenDisplay("")
        screen = XDefaultScreen(dsp)
    else:
        screen = XDefaultScreenOfDisplay(dsp)
    print('X Display %s Screen %s' % (dsp, screen))
    major, minor = GLint(), GLint()
    glXQueryVersion(dsp, major, minor)
    version = (major.value, minor.value)
    print('glX Version: %s.%s' % version)
    if version >= (1, 1):
        print(glXQueryExtensionsString(dsp, screen))
        if version >= (1, 2):
            d = glXGetCurrentDisplay()
            if d:
                print('Current display', d)
            else:
                print('GLX could not find the current display')
        else:
            d = dsp
    if version >= (1, 3):
        elements = GLint(0)
        configs = glXChooseFBConfig(
            dsp, screen, (GLint * len(attributes))(*attributes), elements
        )
        print('%s configs found' % (elements.value))
        for config in range(elements.value):
            print('Config: %s %s' % (config, configs[config][0]))
            samples = ctypes.c_int()
            for attribute in (
                'GLX_FBCONFIG_ID',
                'GLX_BUFFER_SIZE',
                'GLX_LEVEL',
                'GLX_DOUBLEBUFFER',
                'GLX_STEREO',
                'GLX_SAMPLES',
                'GLX_SAMPLE_BUFFERS',
                'GLX_DRAWABLE_TYPE',
            ):
                glXGetFBConfigAttrib(
                    dsp, configs[config], globals()[attribute], samples
                )
                print('%s -> %s' % (attribute, samples.value))
            print()
    from OpenGL.raw.GLX import _types

    print('Extension List', _types.GLXQuerier.getExtensions())


if __name__ == "__main__":
    main()



import basetestcase
import os, logging
import OpenGL
from OpenGL.GL import *
from OpenGL import GLU
try:
    import numpy as np
except ImportError as err:
    np = None
HERE = os.path.abspath(os.path.dirname(__file__))
log = logging.getLogger(__name__)

class TestEvaluators(basetestcase.BaseTest):
    evaluator_ctrlpoints = [[[ -1.5, -1.5, 4.0], [-0.5, -1.5, 2.0], [0.5, -1.5,
        -1.0], [1.5, -1.5, 2.0]], [[-1.5, -0.5, 1.0], [-0.5, -0.5, 3.0], [0.5, -0.5,
        0.0], [1.5, -0.5, -1.0]], [[-1.5, 0.5, 4.0], [-0.5, 0.5, 0.0], [0.5, 0.5,
        3.0], [1.5, 0.5, 4.0]], [[-1.5, 1.5, -2.0], [-0.5, 1.5, -2.0], [0.5, 1.5,
        0.0], [1.5, 1.5, -1.0]]]
    if (not OpenGL.ERROR_ON_COPY) or array:	
        def test_evaluator( self ):
            """Test whether the evaluator functions work"""
            glDisable(GL_CULL_FACE)
            glEnable(GL_MAP2_VERTEX_3)
            glEnable(GL_DEPTH_TEST)
            glEnable(GL_NORMALIZE)
            if np:
                ctrl_points = np.array( self.evaluator_ctrlpoints,'f')
            else:
                ctrl_points = self.evaluator_ctrlpoints
            glMap2f(GL_MAP2_VERTEX_3, 0, 1, 0, 1, ctrl_points)
            glMapGrid2f(20, 0.0, 1.0, 20, 0.0, 1.0)
            glShadeModel(GL_FLAT)
            glEvalMesh2(GL_FILL, 0, 20, 0, 20)
            glTranslatef( 0,0.001, 0 )
            glEvalMesh2(GL_POINT, 0, 20, 0, 20)
    def test_nurbs_raw( self ):
        """Test nurbs rendering using raw API calls"""
        from OpenGL.raw import GLU as GLU_raw
        knots = (GLfloat* 8) ( 0,0,0,0,1,1,1,1 )
        ctlpoints = (GLfloat*(3*4*4))( -3., -3., -3.,
            -3., -1., -3.,
            -3.,  1., -3.,
            -3.,  3., -3.,

        -1., -3., -3.,
            -1., -1.,  3.,
            -1.,  1.,  3.,
            -1.,  3., -3.,

        1., -3., -3.,
             1., -1.,  3.,
             1.,  1.,  3.,
             1.,  3., -3.,

        3., -3., -3.,
            3., -1., -3.,
             3.,  1., -3.,
             3.,  3., -3. )
        theNurb = GLU_raw.gluNewNurbsRenderer()
        GLU_raw.gluBeginSurface(theNurb)
        GLU_raw.gluNurbsSurface(
            theNurb, 
            8, ctypes.byref(knots), 8, ctypes.byref(knots),
            4 * 3, 3, ctypes.byref( ctlpoints ),
            4, 4, GL_MAP2_VERTEX_3
        )
        GLU_raw.gluEndSurface(theNurb)
    if np:
        def test_nurbs_raw_arrays( self ):
            """Test nurbs rendering using raw API calls with arrays"""
            from OpenGL.raw import GLU as GLU_raw 
            knots = np.array( ( 0,0,0,0,1,1,1,1 ), 'f' )
            ctlpoints = np.array( [[[-3., -3., -3.],
                [-3., -1., -3.],
                [-3.,  1., -3.],
                [-3.,  3., -3.]],

            [[-1., -3., -3.],
                [-1., -1.,  3.],
                [-1.,  1.,  3.],
                [-1.,  3., -3.]],

            [[ 1., -3., -3.],
                [ 1., -1.,  3.],
                [ 1.,  1.,  3.],
                [ 1.,  3., -3.]],

            [[ 3., -3., -3.],
                [ 3., -1., -3.],
                [ 3.,  1., -3.],
                [ 3.,  3., -3.]]], 'f' )
            theNurb = GLU_raw.gluNewNurbsRenderer()
            GLU_raw.gluBeginSurface(theNurb)
            GLU_raw.gluNurbsSurface(
                theNurb, 
                8, knots, 8, knots,
                4 * 3, 3, ctlpoints ,
                4, 4, GL_MAP2_VERTEX_3
            )
            GLU_raw.gluEndSurface(theNurb)
        def test_nurbs( self ):
            """Test nurbs rendering"""
            def buildControlPoints( ):
                ctlpoints = np.zeros( (4,4,3), 'f')
                for u in range( 4 ):
                    for v in range( 4):
                        ctlpoints[u][v][0] = 2.0*(u - 1.5)
                        ctlpoints[u][v][1] = 2.0*(v - 1.5);
                        if (u == 1 or u ==2) and (v == 1 or v == 2):
                            ctlpoints[u][v][2] = 3.0;
                        else:
                            ctlpoints[u][v][2] = -3.0;
                return ctlpoints
            controlPoints = buildControlPoints()
            theNurb = GLU.gluNewNurbsRenderer()
            GLU.gluNurbsProperty(theNurb, GLU.GLU_SAMPLING_TOLERANCE, 25.0);
            GLU.gluNurbsProperty(theNurb, GLU.GLU_DISPLAY_MODE, GLU.GLU_FILL);
            knots= np.array ([0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0], "f")
            glPushMatrix();
            try:
                glRotatef(330.0, 1.,0.,0.);
                glScalef (0.5, 0.5, 0.5);

                GLU.gluBeginSurface(theNurb);
                try:
                    GLU.gluNurbsSurface(
                        theNurb,
                        knots, knots,
                        controlPoints,
                        GL_MAP2_VERTEX_3
                    );
                finally:
                    GLU.gluEndSurface(theNurb);
            finally:
                glPopMatrix();
    def test_quadrics( self ):
        """Test for rendering quadric objects"""
        quad = GLU.gluNewQuadric()
        glColor3f( 1,0, 0 )
        GLU.gluSphere( quad, 1.0, 16, 16 )

    if not OpenGL.ERROR_ON_COPY:
        def test_gluNurbsCurve( self ):
            """Test that gluNurbsCurve raises error on invalid arguments"""
            nurb = GLU.gluNewNurbsRenderer()
            GLU.gluBeginCurve( nurb )
            if OpenGL.ERROR_CHECKING:
                self.assertRaises( error.GLUerror,
                    GLU.gluNurbsCurve,
                        nurb, 
                        [0, 1.0],
                        [[0,0,0],[1,0,0],[1,1,0]],
                        GL_MAP1_VERTEX_3,
                )
                self.assertRaises( error.GLUerror,
                    GLU.gluNurbsCurve,
                        nurb, 
                        [],
                        [[0,0,0],[1,0,0],[1,1,0]],
                        GL_MAP1_VERTEX_3,
                )
                self.assertRaises( error.GLUerror,
                    GLU.gluNurbsCurve,
                        nurb, 
                        [],
                        [],
                        GL_MAP1_VERTEX_3,
                )

    def test_gle( self ):
        from OpenGL.GLE import (
            gleSetJoinStyle,
            TUBE_NORM_EDGE, TUBE_JN_ANGLE, TUBE_JN_CAP,
            glePolyCone,
        )
        if (gleSetJoinStyle):
            gleSetJoinStyle(TUBE_NORM_EDGE | TUBE_JN_ANGLE | TUBE_JN_CAP)
            glePolyCone(((-6.0, 6.0, 0.0), (6.0, 6.0, 0.0), (6.0, -6.0, 0.0), (-6.0, -6.0, 0.0), (-6.0, 6.0, 0.0), (6.0, 6.0, 0.0)),
                                ((0.0, 0.0, 0.0), (0.0, 0.8, 0.3), (0.8, 0.3, 0.0), (0.2, 0.3, 0.9), (0.2, 0.8, 0.5), (0.0, 0.0, 0.0)), (1, 1, 3, 0.5, 2, 1))
        else:
            log.warning("No GLE extrusion library")



import pygame, sys
from pygame.locals import *
import OpenGL

OpenGL.USE_ACCELERATE = False
from OpenGL.GL import *
from OpenGL.GL.EXT.framebuffer_object import *
from OpenGL import error


def draw():
    glClearColor(0.0, 0.0, 0.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    # draw stuff here
    pygame.display.flip()


def main():
    pygame.init()
    pygame.display.set_mode((512, 512), OPENGL | DOUBLEBUF)

    # setup a texture
    tex = glGenTextures(1)
    glBindTexture(GL_TEXTURE_2D, tex)
    glTexImage2D(
        GL_TEXTURE_2D, 0, GL_RGBA8, 512, 512, 0, GL_RGBA, GL_UNSIGNED_BYTE, None
    )
    glBindTexture(GL_TEXTURE_2D, 0)

    # setup teh fbo
    fbo = glGenFramebuffersEXT(1)
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo)
    glBindTexture(GL_TEXTURE_2D, tex)

    # this call produces an error!

    glFramebufferTexture2DEXT(
        GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0
    )

    while 1:
        event = pygame.event.poll()
        if event.type is QUIT:
            sys.exit(0)
        draw()
        print('OK')
        break


if __name__ == "__main__":
    main()



#! /usr/bin/env python
"""Tests bug report #32 from github"""
import os
if not os.environ.get( 'PYOPENGL_PLATFORM' ):
    os.environ['PYOPENGL_PLATFORM'] = 'osmesa'
from math import pi, sin, cos
import OpenGL
OpenGL.USE_ACCELERATE = False
from OpenGL.GL import *
from OpenGL.GL.ARB.fragment_program import glGenProgramsARB
from OpenGL.GLU import *
from OpenGL.osmesa import *
width = height = 300

shared_win = None
# Current OSMesa does not seem to support RGB, only RGBA,
# note also the use of the OSMESA_* constants here, they
# seem to match the GL constants, but can't be too careful.
# Also see ./osdemo.py for a working sample of offscreen
# rendering using OSMesa.
ctx = OSMesaCreateContext(OSMESA_RGBA, shared_win)
buf = arrays.GLubyteArray.zeros((height, width, 4))
mesap = arrays.ArrayDatatype.dataPointer(buf)
assert(OSMesaMakeCurrent(ctx, GLuint(mesap), GL_UNSIGNED_BYTE, width, height))


#! /usr/bin/env python
"""Quick script to check for all entry points being defined"""
import re,pprint,logging, json
lmatch= re.compile( r'^\w+[ \t]*\([^)]+\)' )
comment = re.compile( r'([ \t]*\#.*)|passthru\:' )
param = re.compile( r'[ \t]+(?P<type>[^ \t]+)[ \t]+(?P<name>[^ \t]+)([ \t]+(?P<details>.*))?' )
functions = {}
log = logging.getLogger( 'missing' )

def module_for_category( category ):
    """Return the OpenGL.GL.x module for the given category name"""
    if category.startswith( 'VERSION_' ):
        name = 'OpenGL.GL'
    else:
        owner,name = category.split( '_',1)
        if owner.startswith( '3' ):
            owner = owner[1:]
        name = 'OpenGL.GL.%s.%s'%( owner,name )
    return __import__( name, {}, {}, name.split( '.' ))


def main():
    current = None 
    for line in open( 'gl.spec' ):
        line = line.rstrip( '\n' ).rstrip( '\r' )
        if lmatch.match( line ):
            current = {
                '': line.strip(),
                'name': line.strip().split('(')[0],
            }
            functions[current['name']] = current
        elif not current:
            pass
        elif comment.match( line ):
            pass 
        else:
            match = param.match( line )
            if match:
                current.setdefault(match.group('type'),[]).append(
                    (match.group('name'),match.group('details'))
                )
    OUT_MAPPING = {}
    categories = {}
    for key,func in functions.items():
        assert 'category' in func, func
        categories.setdefault(
            func['category'][0][0],
            [] 
        ).append( func )
        out_params = dict([x for x in func.get('param',[]) if 'out' in x[1].split()])
        if out_params:
            OUT_MAPPING['gl'+key] = out_params
    open( 'gl_out_parameters.json','w').write( json.dumps(
        OUT_MAPPING, indent=2,
        sort_keys=True,
    ) )
    for category,funcs in sorted(categories.items()):
        try:
            module = module_for_category( category )
        except ImportError, err:
            log.error( 'Whole category missing: %s', category )
        else:
            log.info( 'Checking category: %s', category )
            for func in funcs:
                funcname = func['name']
                funcname = 'gl'+funcname
                if not hasattr( module, funcname ):
                    log.error(
                        "Missing entry point %s in module %s",
                        funcname,
                        module.__name__,
                    )

if __name__ == "__main__":
    logging.basicConfig()
    main()
